#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Tuple, Union

# 3rd party
import numpy as np
from numpy.typing import NDArray
from xarray.core.dataset import Dataset

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def psd_rebin(
    vdf: Dataset,
    phi: NDArray[np.float32],
    energy0: NDArray[np.float32],
    energy1: NDArray[np.float32],
    esteptable: NDArray[np.uint8],
) -> Tuple[
    NDArray[np.datetime64],
    NDArray[Union[np.float32, np.float64]],
    NDArray[np.float32],
    NDArray[np.float32],
]:
    r"""Convert burst mode distribution into 64 energy channel distribution.

    Takes the burst mode distribution sampled in two energy tables and converts to a
    single energy table with 64 energy channels. Time resolution is halved and phi
    angles are averaged over adjacent times.

    Parameters
    ----------
    vdf : Dataset
        Time series of the particle distribution.
    phi : numpy.ndarray
        Time series of the phi angles.
    energy0 : numpy.ndarray
        Energy table 0.
    energy1 : numpy.ndarray
        Energy table 1.
    esteptable : numpy.ndarray
        Time series of the stepping table between energies (burst).

    Returns
    -------
    tuple
        * time_r : numpy.ndarray
            Time array of the recompiled distribution.
        * vdf_r : numpy.ndarray
            Recompiled distribution.
        * energy_r : numpy.ndarray
            Recompiled energy table.
        * phi_r : numpy.ndarray
            Recompiled phi angles.

    Raises
    ------
    TypeError
        If vdf is not a xarray.Dataset or if phi is not a numpy.ndarray or if energy0 is
        not a numpy.ndarray or if energy1 is not a numpy.ndarray or if esteptable is
        not a numpy.ndarray.

    Notes
    -----
    I'm assuming no gaps in the burst data interval. If there is a gap use
    time_clip before running. To be updated later.

    """
    if not isinstance(vdf, Dataset):
        raise TypeError("vdf must be a xarray.Dataset")

    if not isinstance(phi, np.ndarray):
        raise TypeError("phi must be a numpy.ndarray")

    if not isinstance(energy0, np.ndarray):
        raise TypeError("energy0 must be a numpy.ndarray")

    if not isinstance(energy1, np.ndarray):
        raise TypeError("energy1 must be a numpy.ndarray")

    if not isinstance(esteptable, np.ndarray):
        raise TypeError("step_table must be a numpy.ndarray")

    # Get time and data
    vdf_time: NDArray[np.datetime64] = vdf.time.data
    vdf_data: NDArray[np.float64] = vdf.data.data.astype(np.float64)

    # Sort energy levels
    energy_r: NDArray[np.float32] = np.sort(np.hstack([energy0, energy1]))

    # Define new times
    delta_t: float = np.median(np.diff(vdf_time)).astype(np.int16) / 1e9
    time_r: NDArray[np.datetime64] = vdf_time[:-1:2] + np.timedelta64(
        int(delta_t * 1e9 / 2), "ns"
    )

    # Preallocate output arrays
    vdf_r: NDArray[np.float64] = np.zeros(
        (len(time_r), 64, *vdf.data.shape[2:]), dtype=np.float64
    )
    phi_r: NDArray[np.float32] = np.zeros(
        (len(time_r), vdf.data.shape[2]), dtype=np.float32
    )

    phi_s: NDArray[np.float32] = np.roll(phi, 2, axis=1)
    phi_s[:, 0] = phi_s[:, 0] - 360.0

    time_indices: NDArray[np.int16] = np.arange(0, len(vdf.time) - 1, 2, dtype=np.int16)

    for new_el_num, idx in enumerate(time_indices[:-1]):
        if phi[idx, 0] > phi[idx + 1, 0]:
            phi_r[new_el_num, :] = (phi[idx, :] + phi_s[idx + 1, :]) / 2

            vdf_temp: NDArray[np.float32] = np.roll(
                np.squeeze(vdf_data[idx + 1, ...]),
                2,
                axis=1,
            )

            if esteptable[idx]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf_data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf_data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp

        else:
            phi_r[new_el_num, :] = phi[idx, :] + phi[idx + 1, :]
            phi_r[new_el_num, :] /= 2

            if esteptable[idx]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf_data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf.data.data[idx + 1, ...]
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf_data[idx + 1, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_data[idx, ...]

    out = (time_r, vdf_r.astype(vdf.data.data.dtype), energy_r, phi_r)
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

# Local imports
from ..pyrf.calc_dt import calc_dt
from ..pyrf.convert_fac import convert_fac
from ..pyrf.extend_tint import extend_tint
from ..pyrf.filt import filt
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def lh_wave_analysis(
    tints,
    e_xyz,
    b_scm,
    b_xyz,
    n_e,
    min_freq: float = 10.0,
    max_freq: float = 0.0,
    lowpass_b_xyz: float = 2.0,
):
    r"""Calculates lower-hybrid wave properties from MMS data

    Parameters
    ----------
    tints : list of str
        Time interval
    e_xyz : xarray.DataArray
        Time series pf the electric field
    b_scm : xarray.DataArray
        Time series of the fluctuations of the magnetic field
    b_xyz : xarray.DataArray
        Time series of the background magnetic field
    n_e : xarray.DataArray
        Time series of the number density
    min_freq : float, Optional
        Minimum frequency in the highpass filter for LH fluctuations.
        Default is 10.
    max_freq : float, Optional
        Maximum frequency in the bandpass filter for LH fluctuations.
        Default is 0 (highpass filter).
    lowpass_b_xyz : float, Optional
         Maximum frequency for low-pass filter of background magnetic
         field (FGM)

    Returns
    -------
    phi_eb : xarray.DataArray
        to fill
    v_best : ndarray
        to fill
    dir_best : ndarray
        to fill
    thetas : ndarray
        to fill
    corrs : ndarray
        to fill

    Examples
    --------
    >>> from pyrfu.mms import get_data, lh_wave_analysis

    Define time intervals

    >>> tint_long = ["2015-12-14T01:17:39.000", "2015-12-14T01:17:43.000"]
    >>> tint_zoom = ["2015-12-14T01:17:40.200","2015-12-14T01:17:41.500"]

    Load fields and density

    >>> b_gse = get_data("b_gse_fgm_brst_l2", tint_long, 2)
    >>> e_gse = get_data("e_gse_edp_brst_l2", tint_long, 2)
    >>> b_scm = get_data("b_gse_scm_brst_l2", tint_long, 2)
    >>> n_e = get_data("ne_fpi_brst_l2", tint_long, 2)

    Lower Hybrid Waves Analysis

    >>> opt = dict(lhfilt=[5, 100])
    >>> res = lh_wave_analysis(tint, e_xyz, b_scm, b_xyz, n_e, **opt)

    """

    # Bandpass filter data
    b_xyz = filt(b_xyz, 0, lowpass_b_xyz, 5)
    e_xyz = resample(e_xyz, b_scm)
    n_e = resample(n_e, b_scm)
    b_xyz = resample(b_xyz, b_scm)
    b_scm_fac = convert_fac(b_scm, b_xyz, [1, 0, 0])
    b_scm_fac = filt(b_scm_fac, min_freq, max_freq, 5)
    e_xyz = filt(e_xyz, min_freq, max_freq, 5)

    q_e = constants.elementary_charge
    mu0 = constants.mu_0

    b_mag = np.linalg.norm(b_xyz, axis=1)
    phi_b = b_scm_fac.data[:, 2] * b_mag * 1e-18 / (n_e.data * q_e * mu0 * 1e6)
    phi_b = ts_scalar(b_scm_fac.time.data, phi_b)

    # short buffer so phi_E does not begin at zero.
    tint = extend_tint(tints, [-0.2, 0.2])

    e_xyz = time_clip(e_xyz, tint)
    phi_bs = time_clip(phi_b, tints)

    # Rotate e_xyz into field-aligned coordinates
    b_xyz_tc = time_clip(b_xyz, tints)
    b_mean = np.mean(b_xyz_tc.data, axis=0)
    b_vec = b_mean / np.linalg.norm(b_mean)
    r_temp = [1, 0, 0]
    bxr = np.cross(b_vec, r_temp)
    bxr /= np.linalg.norm(bxr)
    bxrxb = np.cross(bxr, b_vec)

    er1 = e_xyz.data[:, 0] * bxrxb[0]
    er1 += e_xyz.data[:, 1] * bxrxb[1]
    er1 += e_xyz.data[:, 2] * bxrxb[2]

    er2 = e_xyz.data[:, 0] * bxr[0]
    er2 += e_xyz.data[:, 1] * bxr[1]
    er2 += e_xyz.data[:, 2] * bxr[2]

    er3 = e_xyz.data[:, 0] * b_vec[0]
    er3 += e_xyz.data[:, 1] * b_vec[1]
    er3 += e_xyz.data[:, 2] * b_vec[2]

    e_fac = ts_vec_xyz(e_xyz.time.data, np.vstack([er1, er2, er3]).T)

    # Find best direction
    dt_e_fac = calc_dt(e_fac)
    thetas = np.linspace(0, 2 * np.pi, 361)
    corrs = np.zeros(len(thetas))

    for i, theta in enumerate(thetas):
        e_temp = np.cos(theta) * e_fac.data[:, 0]
        e_temp += np.sin(theta) * e_fac.data[:, 1]

        phi_temp = ts_scalar(e_xyz.time.data, np.cumsum(e_temp) * dt_e_fac)
        phi_temp = time_clip(phi_temp, tints)
        phi_temp -= np.mean(phi_temp)

        corrs[i] = np.corrcoef(phi_bs.data, phi_temp.data)

    corrpos = np.argmax(corrs)
    e_best = np.cos(thetas[corrpos]) * e_fac.data[:, 0]
    e_best += np.sin(thetas[corrpos]) * e_fac.data[:, 1]
    e_best = ts_scalar(e_xyz.time.data, e_best)
    phi_best = ts_scalar(e_xyz.time.data, np.cumsum(e_best) * dt_e_fac)
    phi_best = time_clip(phi_best, tints)
    phi_best -= np.mean(phi_best)
    theta_best = thetas[corrpos]
    dir_best = bxrxb * np.cos(theta_best) + bxr * np.sin(theta_best)

    # Find best speed
    # Maximum velocity may need to be increased in rare cases
    vph_vec = np.linspace(1e1, 5e2, 491)
    corr_v = np.zeros(len(vph_vec))

    for i, vph in enumerate(vph_vec):
        phi_e_temp = phi_best.data * vph
        corr_v[i] = np.sum(np.abs(phi_e_temp - phi_bs.data) ** 2)

    corr_vpos = np.argmin(corr_v)
    phi_e_best = phi_best.data * vph_vec[corr_vpos]
    phi_e_best = ts_scalar(phi_bs.time.data, phi_e_best)
    v_best = vph_vec[corr_vpos]

    options = {"coords": [phi_bs.time, ["Ebest", "Bs"]], "dims": ["time", "comp"]}
    phi_eb = xr.DataArray(
        np.vstack([phi_e_best.data, phi_bs.data]).T,
        **options,
    )

    return phi_eb, v_best, dir_best, thetas, corrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import List, Mapping, Optional

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def ts_append(
    inp0: Optional[DataArray] = None, inp1: Optional[DataArray] = None
) -> DataArray:
    r"""Concatenate two time series along the time axis.

    Parameters
    ----------
    inp0 : DataArray, Optional
        Time series of the first input (early times).
    inp1 : DataArray, Optional
        Time series of the second input (late times).

    Returns
    -------
    DataArray
        Concatenated time series.

    Notes
    -----
    The time series must be in the correct time order.

    """

    if inp0 is None:
        out = inp1
    elif inp1 is None:
        out = inp0
    elif inp0 is not None and inp1 is not None:
        out_data = {}

        if inp0.data.ndim != 1:
            out_data["data"] = np.vstack([inp0.data, inp1.data])

        else:
            out_data["data"] = np.hstack([inp0.data, inp1.data])

        out_data["attrs"] = {}

        for k in inp0.attrs:
            if isinstance(inp0.attrs[k], np.ndarray):
                out_data["attrs"][k] = np.hstack([inp0.attrs[k], inp1.attrs[k]])

            else:
                out_data["attrs"][k] = inp0.attrs[k]

        depends: List[Mapping[str, object]] = [{} for _ in range(len(inp0.dims))]

        for i, dim in enumerate(inp0.dims):
            if i == 0 or dim == "time":
                depends[i]["data"] = np.hstack([inp0[dim].data, inp1[dim].data])

                # add attributes
                depends[i]["attrs"] = {}

                for k in inp0[dim].attrs:
                    # if attrs is array time append
                    if isinstance(inp0[dim].attrs[k], np.ndarray):
                        depends[i]["attrs"][k] = np.hstack(
                            [inp0[dim].attrs[k], inp1[dim].attrs[k]],
                        )

                    else:
                        depends[i]["attrs"][k] = inp0[dim].attrs[k]

            else:
                # Use values of other coordinates of inp0 assuming equal to inp1
                depends[i]["data"] = inp0[dim].data

                # add attributes
                depends[i]["attrs"] = {}

                for k in inp0[dim].attrs:
                    depends[i]["attrs"][k] = inp0[dim].attrs[k]

        # Create DataArray
        out = xr.DataArray(
            out_data["data"],
            coords=[depend["data"] for depend in depends],
            dims=inp0.dims,
            attrs=out_data["attrs"],
        )

        for i, dim in enumerate(out.dims):
            out[dim].attrs = depends[i]["attrs"]

    else:
        raise ValueError("At least one input must be provided")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def calc_fs(inp: Union[Dataset, DataArray]) -> float:
    r"""Computes the sampling frequency of the input time series.

    Parameters
    ----------
    inp : DataArray or Dataset
        Time series of the input variable.

    Returns
    -------
    f_samp : float
        Sampling frequency in Hz.

    """
    # Check input type
    if not isinstance(inp, (Dataset, DataArray)):
        raise TypeError("Input must be a time series")

    # Convert time to nanoseconds
    time_datetime64 = inp.time.data.astype(np.datetime64)
    time_ns = time_datetime64.astype(np.int64)

    # Calculate the sampling frequency
    dt_ns = np.diff(time_ns)
    f_samp = 1e9 / np.median(dt_ns)

    return f_samp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt

# 3rd party imports
import numpy as np
from matplotlib import colors

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_reduced_2d(ax, f2d, clim: list = None):
    r"""Plot the 2D recuded distribution `f2d` onto the axis `ax`.

    Parameters
    ----------
    ax : matplotlib.axes._subplots.AxesSubplot
        Axis.
    f2d : xarray.DataArray
        2D reduced distribution.
    clim : list of floats, Optional
        Coloraxis limits.

    Returns
    -------
    ax : matplotlib.axes._subplots.AxesSubplot
        Axis.
    cax : matplotlib.axes._axes.Axes
        Colorbar axis.

    """

    if clim is None:
        clim = [1e-13, 1e-8]

    clim_log = np.log10(clim)
    n_levs = int(np.round(clim_log[1]) - np.round(clim_log[0])) + 1

    im = ax.pcolormesh(
        f2d.vx.data,
        f2d.vy.data,
        np.transpose(f2d.data),
        norm=colors.LogNorm(vmin=clim[0], vmax=clim[1]),
        cmap="Spectral_r",
        rasterized=True,
        shading="auto",
    )

    c_lines = ax.contour(
        f2d.vx.data,
        f2d.vy.data,
        np.transpose(f2d.data),
        levels=np.logspace(clim_log[0], clim_log[1], n_levs),
        norm=colors.LogNorm(vmin=clim[0], vmax=clim[1]),
        cmap="viridis",
    )

    vx_lim = np.max(abs(f2d.vx.data))
    vy_lim = np.max(abs(f2d.vy.data))

    ax.set_xlim(np.array([-vx_lim, vx_lim]))
    ax.set_ylim(np.array([-vy_lim, vy_lim]))
    ax.set_aspect("equal")

    f = plt.gcf()
    pos = ax.get_position()
    cax = f.add_axes([pos.x0, pos.y0 + pos.height + 0.01, pos.width, 0.01])
    cbar = f.colorbar(mappable=im, cax=cax, ax=ax, orientation="horizontal")
    cbar.add_lines(c_lines)
    cax.xaxis.tick_top()
    cax.xaxis.set_label_position("top")

    return ax, cax



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

# Built-in imports
import os

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def db_init(local_data_dir):
    r"""Setup the default path of SolO data.

    Parameters
    ----------
    local_data_dir : str
        Path to the data.

    """

    # Normalize the path and make sure that it exists
    local_data_dir = os.path.normpath(local_data_dir)
    assert os.path.exists(
        local_data_dir,
    ), f"{local_data_dir} doesn't exists!!"

    # Path to the configuration file.
    pkg_path = os.path.dirname(os.path.abspath(__file__))

    # Read the current version of the configuration
    with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
        config = json.load(fs)

    # Overwrite the configuration file with the new path
    with open(os.path.join(pkg_path, "config.json"), "w", encoding="utf-8") as fs:
        config["local_data_dir"] = local_data_dir
        json.dump(config, fs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Optional

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from scipy import signal
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def medfilt(inp: DataArray, kernel_size: Optional[int] = None) -> DataArray:
    r"""Applies a median filter over npts points to inp.

    Parameters
    ----------
    inp : DataArray
        Time series of the input variable.
    kernel_size : int, Optional
        Number of points of median filter. Default is a 3-point median filter.

    Returns
    -------
    DataArray
        Time series of the median filtered input variable.

    Raises
    ------
    TypeError
        If inp is not a DataArray.
    ValueError
        If inp is not 1D, 2D or 3D.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_list = numpy.arange(1,5)

    Load magnetic field and electric field

    >>> r_mms, b_mms = [[] * 4 for _ in range(2)]
    >>> for mms_id in range(1, 5):
    >>> 	r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>

    Compute current density, etc

    >>> j_xyz, _, b_xyz, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    Get J sampling frequency

    >>> fs = pyrf.calc_fs(j_xyz)

    Median filter over 1s

    >>> j_xyz = pyrf.medfilt(j_xyz,fs)

    """

    # Check input
    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a DataArray")

    # Get number of times
    n_times: int = len(inp)

    # Check kernel size
    if kernel_size is None:
        # Set default kernel size to 3
        kernel_size = 3
    elif kernel_size % 2 == 0:
        # If kernel size is even, add 1.
        kernel_size += 1

    # Check if input is 1D, 2D or 3D
    if inp.ndim == 1:
        # Add a dimension to the input if it is 1D
        inp_data: NDArray[np.float32] = inp.data[:, np.newaxis]
    elif inp.ndim == 2:
        # Keep input as is if it is 2D
        inp_data = inp.data
    elif inp.ndim == 3:
        # Reshape input if it is 3D to 2D (n_times, 9)
        inp_data = np.reshape(inp.data, [n_times, 9])
    else:
        raise ValueError("Input must be 1D, 2D or 3D")

    # Preallocate output
    out_data: NDArray[np.float32] = np.zeros(inp_data.shape, dtype=np.float32)

    # Apply median filter
    for i in range(inp_data.shape[1]):
        out_data[:, i] = signal.medfilt(inp_data[:, i], kernel_size)

    # Reshape output if input was 3D
    if inp_data.shape[1] == 9:
        out_data = np.reshape(out_data, [n_times, 3, 3])

    # Create output DataArray
    out: DataArray = xr.DataArray(
        np.squeeze(out_data), coords=inp.coords, dims=inp.dims, attrs=inp.attrs
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime
import json
import os
import re
from typing import Any, Mapping, Optional, Union

# 3rd party imports
import boto3
import numpy as np
from dateutil import parser
from dateutil.rrule import DAILY, rrule

from pyrfu.mms.db_init import MMS_CFG_PATH

# Local imports
from pyrfu.pyrf.datetime642iso8601 import datetime642iso8601
from pyrfu.pyrf.iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def list_files_aws(
    tint: list[str],
    mms_id: Union[str, int],
    var: Mapping[str, str],
    bucket_prefix: Optional[str] = "",
) -> list[dict[str, Any]]:
    r"""List files from Amazon Web Services (AWS).

    Find available files from the Amazon Wed Services (AWS) for the target instrument,
    data type, data rate, mms_id and level during the target time interval.

    Parameters
    ----------
    tint : list of str
        Time interval
    mms_id : str or int
        Index of the spacecraft
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument
            * var["tmmode"] : data rate
            * var["lev"] : data level
            * var["dtype"] : data type
    bucket_prefix : str, Optional
        Name of AWS S3 bucket.

    Returns
    -------
    file_names : list of str
        List of files corresponding to the parameters in the selected time
        interval

    Raises
    ------
    FileNotFoundError
        If the path doesn't exist in the AWS S3 bucket or if the bucket doesn't exist.
    TypeError
        If the time interval is not array_like or if tint values are not in datetime64
        or str.

    """
    # Start S3 session
    s3 = boto3.resource("s3")

    # Check path
    if not bucket_prefix:
        # Read the current version of the MMS configuration file
        with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
            config = json.load(fs)

        aws_path_split = config["aws"].split("/")
    else:
        aws_path_split = bucket_prefix.split("/")

    bucket_name, prefix = aws_path_split[0], "/".join(aws_path_split[1:])

    # Make sure that the data path exists
    bucket = s3.Bucket(bucket_name)

    if not bucket:
        raise FileNotFoundError(f"{bucket_name} doesn't exist!!")

    if bucket.objects.filter(Prefix=prefix):
        raise FileNotFoundError(f"{prefix} doesn't exist in {bucket_name}")

    # Check time interval
    if isinstance(tint, list):
        tint_array = np.array(tint)
    else:
        raise TypeError("tint must be a list!!")

    # Convert time interval to ISO 8601
    if isinstance(tint_array[0], str):
        tint_iso8601 = datetime642iso8601(iso86012datetime64(tint_array))
    else:
        raise TypeError("Values must be in str!!")

    if not isinstance(mms_id, str):
        mms_id = str(mms_id)

    # directory and file name search patterns:
    # - assume directories are of the form:
    # (srvy, SITL): spacecraft/instrument/rate/level[/datatype]/year/month/
    # (brst): spacecraft/instrument/rate/level[/datatype]/year/month/day/
    # - assume file names are of the form:
    # spacecraft_instrument_rate_level[_datatype]_YYYYMMDD[hhmmss]_version.cdf

    file_name = (
        f"mms{mms_id}_{var['inst']}_{var['tmmode']}_{var['lev']}"
        + r"(_)?.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf"
    )

    d_start = parser.parse(parser.parse(tint_iso8601[0]).strftime("%Y-%m-%d"))
    until_ = parser.parse(tint_iso8601[1]) - datetime.timedelta(seconds=1)
    days = rrule(DAILY, dtstart=d_start, until=until_)

    if var["dtype"] == "" or var["dtype"] is None:
        level_and_dtype = var["lev"]
    else:
        level_and_dtype = os.sep.join([var["lev"], var["dtype"]])

    files_out = []

    for date in days:
        if var["tmmode"] == "brst":
            bucket_prefix = os.sep.join(
                [
                    prefix,
                    f"mms{mms_id}",
                    var["inst"],
                    var["tmmode"],
                    level_and_dtype,
                    date.strftime("%Y"),
                    date.strftime("%m"),
                    date.strftime("%d"),
                ],
            )
        else:
            bucket_prefix = os.sep.join(
                [
                    prefix,
                    f"mms{mms_id}",
                    var["inst"],
                    var["tmmode"],
                    level_and_dtype,
                    date.strftime("%Y"),
                    date.strftime("%m"),
                ],
            )

        full_path = os.sep.join([re.escape(bucket_prefix), file_name])

        regex = re.compile(full_path)

        files = bucket.objects.filter(Prefix=bucket_prefix)

        for file in files:
            this_file = file.key
            matches = regex.match(this_file)
            if matches:
                this_time = parser.parse(matches.groups()[1])
                if d_start <= this_time <= until_:
                    if this_file not in files_out:
                        files_out.append(
                            {
                                "s3_obj": file,
                                "timetag": "",
                                "full_name": this_file,
                                "file_size": "",
                            },
                        )

    return files_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _check_time(proton_phxtof, proton_extof):
    data_size = [len(proton_phxtof), len(proton_extof)]

    if data_size[0] == data_size[1]:
        # identify mismatching timesteps
        cond = proton_phxtof.time.data != proton_extof.time.data
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            extof_inds = []
            phxtof_inds = []
            for i_t in range(data_size[1]):
                dt_dummy = np.min(
                    np.abs(
                        proton_extof.time.data[i_t] - proton_extof.time.data,
                    ),
                )
                t_ind = np.argmin(
                    np.abs(
                        proton_extof.time.data[i_t] - proton_extof.time.data,
                    ),
                )
                if dt_dummy == 0:
                    extof_inds.append(i_t)
                    if not phxtof_inds:
                        phxtof_inds.append(i_t)
                    else:
                        phxtof_inds.append(t_ind)

            time_data = proton_extof.time.data[extof_inds]
            phxtof_data = proton_phxtof.data[phxtof_inds, ...]
            extof_data = proton_extof.data[extof_inds, ...]
        else:
            time_data = proton_extof.time.data
            phxtof_data = proton_phxtof.data
            extof_data = proton_extof.data

    elif data_size[0] > data_size[1]:
        cond = proton_phxtof.time.data[: data_size[1]] != proton_extof.time.data
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            extof_inds = []
            phxtof_inds = []
            for i_t in range(data_size[1]):
                dt_dummy = np.min(
                    np.abs(
                        proton_extof.time.data[i_t] - proton_phxtof.time.data,
                    ),
                )
                t_ind = np.argmin(
                    np.abs(
                        proton_extof.time.data[i_t] - proton_phxtof.time.data,
                    ),
                )
                if dt_dummy == 0:
                    extof_inds.append(i_t)
                    if not phxtof_inds:
                        phxtof_inds.append(i_t)
                    else:
                        phxtof_inds.append(t_ind)

            time_data = proton_extof.time.data[extof_inds]
            phxtof_data = proton_phxtof.data[phxtof_inds, ...]
            extof_data = proton_extof.data[extof_inds, ...]
        else:
            time_data = proton_extof.time.data
            phxtof_data = proton_phxtof.data[: data_size[1], ...]
            extof_data = proton_extof.data

    elif data_size[0] < data_size[1]:
        cond = proton_phxtof.time.data != proton_extof.time.data[: data_size[0]]
        bad_inds = np.where(cond)[0]

        if bad_inds.size:
            extof_inds = []
            phxtof_inds = []
            for i_t in range(data_size[0]):
                dt_dummy = np.min(
                    np.abs(
                        proton_phxtof.time.data[i_t] - proton_phxtof.time.data,
                    ),
                )
                t_ind = np.argmin(
                    np.abs(
                        proton_phxtof.time.data[i_t] - proton_phxtof.time.data,
                    ),
                )
                if dt_dummy == 0:
                    phxtof_inds.append(i_t)
                    if not phxtof_inds:
                        extof_inds.append(i_t)
                    else:
                        extof_inds.append(t_ind)

            time_data = proton_phxtof.time.data[phxtof_inds]
            phxtof_data = proton_phxtof.data[phxtof_inds, ...]
            extof_data = proton_extof.data[extof_inds, ...]

        else:
            time_data = proton_phxtof.time.data
            phxtof_data = proton_phxtof.data
            extof_data = proton_extof.data[: data_size[0], ...]

    else:
        raise ValueError

    return time_data, phxtof_data, extof_data


def _combine_attrs(attrs1, attrs2):
    attrs = {}
    for k in attrs1:
        if k.lower() == "global":
            attrs[k] = _combine_attrs(attrs1[k], attrs2[k])
        elif k not in ["delta_energy_plus", "delta_energy_minus"]:
            if attrs1[k] == attrs2[k]:
                attrs[k] = attrs1[k]
            else:
                attrs[k] = [attrs1[k], attrs2[k]]
        else:
            continue

    return attrs


def eis_combine_proton_spec(phxtof_allt, extof_allt):
    r"""Combine ExTOF and PHxTOF proton energy spectra into a single combined
    Dataset.

    Parameters
    ----------
    phxtof_allt : xarray.Dataset
        Dataset containing the PHxTOF energy spectrum of the 6 telescopes.
    extof_allt : xarray.Dataset
        Dataset containing the ExTOF energy spectrum of the 6 telescopes.

    Returns
    -------
    comb_allt : xarray.Dataset
        Dataset containing the combined PHxTOF and ExTOF energy spectrum of
        the 6 telescopes.

    """

    scopes_phxtof = list(filter(lambda x: x[0] == "t", phxtof_allt))
    scopes_extof = list(filter(lambda x: x[0] == "t", extof_allt))
    assert scopes_extof == scopes_phxtof

    # Get energy deltas for PHxTOF and ExTOF
    delta_energy_plus_phxtof = phxtof_allt.attrs["delta_energy_plus"]
    delta_energy_minus_phxtof = phxtof_allt.attrs["delta_energy_minus"]

    delta_energy_plus_extof = extof_allt.attrs["delta_energy_plus"]
    delta_energy_minus_extof = extof_allt.attrs["delta_energy_minus"]

    out_keys = list(filter(lambda x: x not in scopes_extof, extof_allt))
    out_dict = {k: extof_allt[k] for k in out_keys if k != "sector"}

    energy_combined_low, energy_combined_hig = [None] * 2

    time_sect, _, extof_sect = _check_time(
        phxtof_allt["sector"],
        extof_allt["sector"],
    )
    sect = xr.DataArray(
        extof_sect,
        coords=[time_sect],
        dims=["time"],
        attrs=extof_allt["sector"].attrs,
    )

    _, _, extof_spin = _check_time(phxtof_allt["spin"], extof_allt["spin"])

    spin = xr.DataArray(
        extof_spin,
        coords=[time_sect],
        dims=["time"],
        attrs=extof_allt["spin"].attrs,
    )

    for scope in scopes_phxtof:
        proton_phxtof = phxtof_allt[scope]
        proton_extof = extof_allt[scope]

        time_data, phxtof_data, extof_data = _check_time(
            proton_phxtof,
            proton_extof,
        )

        energy_phxtof = proton_phxtof.energy.data
        energy_extof = proton_extof.energy.data

        idx_phxtof = np.where(energy_phxtof < energy_extof[0])[0]
        cond_ = np.logical_and(
            energy_phxtof > energy_extof[0],
            energy_phxtof < energy_phxtof[-1],
        )
        idx_phxtof_cross = np.where(cond_)[0]

        idx_extof_cross = np.where(energy_extof < energy_phxtof[-2])[0]
        idx_extof = np.where(energy_extof > energy_phxtof[-2])[0]

        n_phxtof = idx_phxtof.size
        n_phxtof_cross = idx_phxtof_cross.size
        n_extof = idx_extof.size

        n_en = n_phxtof + n_phxtof_cross + n_extof

        energy_combined = np.zeros(n_en)
        energy_combined_low, energy_combined_hig = [np.zeros(n_en) for _ in range(2)]

        data_combined = np.zeros((len(time_data), n_en))
        data_combined[:, :n_phxtof] = phxtof_data[:, idx_phxtof]
        energy_combined[:n_phxtof] = energy_phxtof[idx_phxtof]
        energy_combined_low[:n_phxtof] = (
            energy_combined[:n_phxtof] - delta_energy_minus_phxtof[idx_phxtof]
        )
        energy_combined_hig[:n_phxtof] = (
            energy_combined[:n_phxtof] + delta_energy_plus_phxtof[idx_phxtof]
        )

        for (i, i_phx), i_ex in zip(
            enumerate(idx_phxtof_cross),
            idx_extof_cross,
        ):
            idx_ = n_phxtof + i
            data_combined[:, idx_] = np.nanmean(
                np.vstack([phxtof_data[:, i_phx], extof_data[:, i_ex]]),
                axis=0,
            )
            energy_combined_low[idx_] = np.nanmin(
                [
                    energy_phxtof[idx_] - delta_energy_minus_phxtof[idx_],
                    energy_extof[i] - delta_energy_minus_extof[i],
                ],
            )
            energy_combined_hig[idx_] = np.nanmax(
                [
                    energy_phxtof[idx_] + delta_energy_plus_phxtof[idx_],
                    energy_extof[i] + delta_energy_plus_extof[i],
                ],
            )
            energy_combined[idx_] = np.sqrt(
                energy_combined_low[idx_] * energy_combined_hig[idx_],
            )

        data_combined[:, -n_extof:] = extof_data[:, idx_extof]
        energy_combined[-n_extof:] = energy_extof[idx_extof]
        energy_combined_low[-n_extof:] = (
            energy_extof[idx_extof] - delta_energy_minus_extof[idx_extof]
        )
        energy_combined_hig[-n_extof:] = (
            energy_extof[idx_extof] + delta_energy_plus_extof[idx_extof]
        )

        attrs = _combine_attrs(
            phxtof_allt[scope].attrs,
            extof_allt[scope].attrs,
        )

        out_dict[scope] = xr.DataArray(
            data_combined,
            coords=[time_data, energy_combined],
            dims=["time", "energy"],
            attrs=attrs,
        )

    out_dict["sector"] = sect
    out_dict["spin"] = spin

    # Combine attributes for all telescopes and set energy deltas as attributes
    attrs = _combine_attrs(phxtof_allt.attrs, extof_allt.attrs)
    attrs = {
        "delta_energy_minus": energy_combined_low,
        "delta_energy_plus": energy_combined_hig,
        **attrs,
    }

    # Create Dataset
    comb_allt = xr.Dataset(out_dict, attrs=attrs)
    comb_allt.time.attrs = _combine_attrs(
        phxtof_allt.time.attrs,
        extof_allt.time.attrs,
    )
    comb_allt.energy.attrs = _combine_attrs(
        phxtof_allt.energy.attrs,
        extof_allt.energy.attrs,
    )

    return comb_allt



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime
from typing import Union

# 3rd party imports
import numpy as np
from numpy.typing import NDArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def iso86012datetime(
    time: Union[list[str], NDArray[np.str_]],
) -> list[datetime.datetime]:
    r"""Convert ISO 8601 time to datetime.

    Parameters
    ----------
    time : array_like
        Time

    Returns
    -------
    time_datetime : list of datetime.datetime
        Time in datetime format.

    """
    # Check input type
    if isinstance(time, list):
        time_array = np.array(time)
    elif isinstance(time, np.ndarray):
        time_array = time
    else:
        raise TypeError("time must be list or numpy.ndarray")

    # Make sure that str is in ISO8601 format
    time_datetime64 = time_array.astype("datetime64[ns]")
    time_iso8601 = time_datetime64.astype(str)

    # ISO 8601 format with miliseconds precision (max precision for datetime)
    fmt = "%Y-%m-%dT%H:%M:%S.%f"

    # Convert to datetime format
    time_datetime = [datetime.datetime.strptime(t_[:-3], fmt) for t_ in time_iso8601]

    return time_datetime



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

from .cross import cross
from .normalize import normalize

# Local imports
from .resample import resample
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def mean(inp, r_xyz, b_xyz, dipole_axis: xr.DataArray = None):
    r"""Put inp into mean field coordinates defined by position vector r and
    magnetic field b if earth magnetic dipole axis z is given then  uses
    another algorithm (good for auroral passages)

    Parameters
    ----------
    inp : xarray.DataArray
        Input field to put into MF coordinates.
    r_xyz : xarray.DataArray
        Time series of the spacecraft position.
    b_xyz : xarray.DataArray
        Time series of the background magnetic field.
    dipole_axis : xarray.DataArray, Optional
        Earth magnetic dipole axis.

    Returns
    -------
    out : xarray.DataArray
        Input field in mean field coordinates.

    """

    if dipole_axis is not None:
        assert isinstance(dipole_axis, xr.DataArray)
        flag_dipole = True

        dipole_axis = resample(dipole_axis, inp)

    else:
        flag_dipole = False

    # Make sure that spacecraft position and magnetic field sampling matches
    # input sampling
    r_xyz = resample(r_xyz, inp)
    b_xyz = resample(b_xyz, inp)

    b_hat = normalize(b_xyz)

    if not flag_dipole:
        bxr = cross(b_hat, r_xyz)
        bxr /= np.linalg.norm(bxr, axis=1)[:, None]
    else:
        fact = -1 * np.ones(len(b_xyz))
        fact[np.sum(b_xyz * r_xyz) > 0] = 1
        bxr = np.cross(dipole_axis, b_xyz) * fact[:, None]
        bxr /= np.linalg.norm(bxr, axis=1)[:, None]

    bxrxb = np.cross(bxr, b_hat)

    out_data = np.zeros(inp.data.shape)
    out_data[:, 0] = np.sum(bxrxb * inp, axis=1)
    out_data[:, 1] = np.sum(bxr * inp, axis=1)
    out_data[:, 2] = np.sum(b_hat * inp, axis=1)

    out = ts_vec_xyz(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Mapping, Optional, Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def ts_tensor_xyz(
    time: NDArray[np.datetime64],
    data: NDArray[Union[np.float32, np.float64]],
    attrs: Optional[Mapping[str, object]] = None,
) -> DataArray:
    r"""Create a time series containing a 2nd order tensor.

    Parameters
    ----------
    time : numpy.ndarray
        Array of times.
    data : numpy.ndarray
        Data corresponding to the time list.
    attrs : dict, Optional
        Attributes of the data list.

    Returns
    -------
    DataArray
        2nd order tensor time series.

    Raises
    ------
    TypeError
        If time or data is not a numpy.ndarray.
    ValueError
        * If time and data do not have the same length.
        * If data does not have shape (n, 3, 3).

    """

    # Check input type
    if not isinstance(time, np.ndarray):
        raise TypeError("time must be a numpy.ndarray")

    if not isinstance(data, np.ndarray):
        raise TypeError("data must be a numpy.ndarray")

    # Check data and time have the same length
    if len(time) != len(data):
        raise ValueError("Time and data must have the same length")

    # Check input shape must be (n, 3, 3)
    if data.ndim != 3 or data.shape[1:] != (3, 3):
        raise ValueError("data must have shape (n, 3, 3)")

    if attrs is None or not isinstance(attrs, dict):
        attrs = {}

    out = xr.DataArray(
        data,
        coords=[time[:], ["x", "y", "z"], ["x", "y", "z"]],
        dims=["time", "rcomp", "ccomp"],
        attrs=attrs,
    )

    out.attrs["TENSOR_ORDER"] = 2

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib as mpl
import matplotlib.pyplot as plt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plot_spectr(
    axis,
    inp,
    yscale: str = "linear",
    cscale: str = "linear",
    clim: list = None,
    cmap: str = None,
    colorbar: str = "right",
    aspect: str = "auto",
    **kwargs,
):
    r"""Plot a spectrogram using pcolormesh.

    Parameters
    ----------
    axis : matplotlib.pyplot.subplotsaxes
        Target axis to plot. If None create a new figure.
    inp : xarray.DataArray
        Input 2D data to plot.
    yscale : {"linear", "log"}, Optional
        Y-axis scaling. Default is "" (linear).
    cscale : {"linear", "log"}, Optional
        C-axis scaling. Default is "" (linear).
    clim : list, Optional
        C-axis bounds. Default is None (autolim).
    cmap : str, Optional
        Colormap. Default is "jet".
    colorbar : str, Optional
        Location of the colorbar with respect to the axis.
        Set to "none" to hide.

    Other Parameters
    ----------------
    **kwargs
        Keyword arguments.

    Returns
    -------
    fig : figure
        Figure with axis.
    axs : matplotlib.pyplot.subplotsaxes
        Axis with spectrum.
    caxs : matplotlib.pyplot.subplotsaxes
        Only if colorbar is True.

    """

    if axis is None:
        fig, axis = plt.subplots(1)
    else:
        fig = plt.gcf()

    if not cmap or isinstance(cmap, str):
        cmap = mpl.colormaps.get_cmap(cmap)

    # else:
    #    raise TypeError(
    #        "cmap must be a string. "
    #        "To add a custom colormap use mpl.colormaps.register(custom)."
    #    )

    if cscale == "log":
        if clim is not None and isinstance(clim, list):
            options = {
                "norm": mpl.colors.LogNorm(vmin=clim[0], vmax=clim[1]),
                "cmap": cmap,
            }
        else:
            options = {"norm": mpl.colors.LogNorm(), "cmap": cmap}
    elif cscale == "linear":
        if clim is not None and isinstance(clim, list):
            options = {"cmap": cmap, "vmin": clim[0], "vmax": clim[1]}
        else:
            options = {"cmap": cmap, "vmin": None, "vmax": None}
    else:
        options = {"norm": cscale, "cmap": cmap}

    x_data, y_data = [inp.coords[inp.dims[0]], inp.coords[inp.dims[1]]]

    image = axis.pcolormesh(
        x_data.data,
        y_data.data,
        inp.data.T,
        rasterized=True,
        shading="auto",
        **options,
    )

    if x_data.dtype == "<M8[ns]":
        locator = mpl.dates.AutoDateLocator(minticks=3, maxticks=7)
        formatter = mpl.dates.ConciseDateFormatter(locator)
        axis.xaxis.set_major_locator(locator)
        axis.xaxis.set_major_formatter(formatter)

    if yscale == "log":
        axis.set_yscale("log")
        axis.yaxis.set_major_locator(mpl.ticker.LogLocator(base=10.0, numticks=4))

    axis.set_axisbelow(False)
    axis.set_ylim(inp[inp.dims[1]].data[[0, -1]])
    axis.set_aspect(aspect)

    if colorbar.lower() == "right":
        if kwargs.get("pad"):
            pad = kwargs["pad"]
        else:
            pad = 0.01

        pos = axis.get_position()
        cax = fig.add_axes(
            [pos.x0 + pos.width + pad, pos.y0, 0.01, pos.height],
        )
        plt.colorbar(mappable=image, cax=cax, ax=axis, orientation="vertical")

        cax.yaxis.set_ticks_position(colorbar.lower())
        cax.yaxis.set_label_position(colorbar.lower())

        cax.set_axisbelow(False)

        if cscale == "log":
            cax.yaxis.set_major_locator(
                mpl.ticker.LogLocator(base=10.0, numticks=4),
            )
        else:
            cax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(4))

        out = (axis, cax)
    elif colorbar.lower() == "top":
        if kwargs.get("pad"):
            pad = kwargs["pad"]
        else:
            pad = 0.01

        pos = axis.get_position()
        cax = fig.add_axes(
            [pos.x0, pos.y0 + pos.height + pad, pos.width, 0.01],
        )
        plt.colorbar(
            mappable=image,
            cax=cax,
            ax=axis,
            orientation="horizontal",
        )

        cax.xaxis.set_ticks_position(colorbar.lower())
        cax.xaxis.set_label_position(colorbar.lower())

        cax.set_axisbelow(False)

        if cscale == "log":
            cax.xaxis.set_major_locator(
                mpl.ticker.LogLocator(base=10.0, numticks=4),
            )
        else:
            cax.xaxis.set_major_locator(mpl.ticker.MaxNLocator(4))

        out = (axis, cax)
    elif colorbar.lower() == "none":
        out = axis
    else:
        raise NotImplementedError("colorbar must be: 'right', 'top', or 'none'")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyrfu import dispersion, lp, maven, mms, models, plot, pyrf

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2026"
__license__ = "MIT"
__version__ = "2.4.21"
__status__ = "Prototype"

__all__ = ["dispersion", "lp", "maven", "mms", "models", "plot", "pyrf"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import os
import shutil
from typing import Mapping, Optional, Union

# Local imports
from pyrfu.mms.db_init import MMS_CFG_PATH
from pyrfu.mms.list_files import list_files

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def copy_files(
    var: Mapping[str, str],
    tint: list[str],
    mms_id: Union[int, str],
    tar_path: str,
    data_path: Optional[str] = "",
) -> None:
    r"""Copy files from local as defined in config.json to the target path.

    Parameters
    ----------
    var : dict
        Dictionary containing 4 keys
            * var["inst"] : name of the instrument.
            * var["tmmode"] : data rate.
            * var["lev"] : data level.
            * var["dtype"] : data type.
    tint : list
        Time interval.
    mms_id : int or str
        Index of the spacecraft.
    tar_path : str
        Target path to put files.
    data_path : str, Optional
        Local path to MMS data. Default uses that provided in pyrfu.mms.config.json

    """
    # Normalize the target path and make sure it exists.
    tar_path = os.path.normpath(tar_path)
    assert os.path.exists(tar_path), f"{tar_path} doesn't exist!!"

    if not data_path:
        # Read the current version of the MMS configuration file
        with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
            config = json.load(fs)

        normed_path = os.path.normpath(config["local"])
    else:
        normed_path = os.path.normpath(data_path)

    # Make sure the local path exists.
    assert os.path.exists(normed_path), f"{normed_path} doesn't exist!!"

    # List files that matches the requirements (instrument, date level,
    # data type, data rate) in the time interval for the target spacecraft.
    files = list_files(tint, mms_id, var, data_path=normed_path)

    for file in files:
        # Make paths
        relative_path = os.path.relpath(file, normed_path)
        directory_path = os.path.join(tar_path, os.path.dirname(relative_path))
        target_file = os.path.join(directory_path, os.path.basename(file))

        # Create directories in target path
        os.makedirs(directory_path, exist_ok=True)

        # Copy file
        shutil.copy(file, target_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

mass_and_charge = {
    "hydrogen+": [1.04535e-2, 1],
    "helium+": [4.18138e-2, 1],
    "helium++": [4.18138e-2, 2],
    "oxygen+": [0.167255, 1],
    "oxygen++": [0.167255, 2],
}


def _get_energy(inp, dim):
    energy_len = len(inp.ccomp.data)
    energy_reform = np.reshape(inp.ccomp.data, [energy_len, 1, 1])
    # repeated across theta
    energy_rebin1 = np.repeat(energy_reform, dim[1], axis=2)
    out_energy = np.repeat(energy_rebin1, dim[2], axis=1)

    return out_energy


def _get_theta(inp, azimuth_dim):
    energy_len = len(inp.ccomp.data)

    # elevations are constant across time
    # convert colat -> lat
    theta_len = len(inp.rcomp.data)
    theta_reform = 90.0 - np.reshape(inp.rcomp.data, [1, 1, theta_len])

    # in the IDL code, we use reform to repeat the vector above
    # here, we'll do the same thing with np.repeat
    # repeated across phi
    theta_rebin1 = np.repeat(theta_reform, azimuth_dim[1], axis=1)
    # repeated across energy
    out_theta = np.repeat(theta_rebin1, energy_len, axis=0)

    return out_theta


def _get_phi(azimuth, full):
    # get azimuth
    # -shift from time-azi.-elev.-en. to time-en.-azi.-elev.
    out_phi = azimuth.data[full, :, :, :]

    if out_phi.ndim == 4:
        out_phi = out_phi.transpose([0, 3, 1, 2])
    elif out_phi.ndim == 3:
        out_phi = out_phi.transpose([2, 0, 1])

    return out_phi


def _get_dphi(azimuth, full, out_phi, inp):
    energy_len = len(inp.ccomp.data)
    theta_len = len(inp.rcomp.data)
    phi_len = azimuth.shape[1]
    # get dphi
    #  - use median distance between subsequent phi measurements within each
    #  distribution (median is used to discard large differences across 0=360)
    #  - preserve dimensionality in case differences arise across energy or
    #  elevation

    if out_phi.ndim == 4:
        out_dphi = np.median(np.diff(azimuth.data[full, ...], axis=1), axis=1)
        out_dphi = np.transpose(out_dphi, [0, 2, 1])
        dphi_reform = np.reshape(
            out_dphi,
            [full.size, energy_len, theta_len, 1],
        )
        out_dphi = np.repeat(dphi_reform, phi_len, axis=3)
    elif out_phi.ndim == 3:
        out_dphi = np.median(np.diff(azimuth.data[full, ...], axis=1), axis=0)
        out_dphi = out_dphi.transpose([1, 0])
        dphi_reform = np.reshape(out_dphi, [energy_len, theta_len, 1])
        out_dphi = np.repeat(dphi_reform, phi_len, axis=2)
    else:
        raise TypeError("Invalid shape of out_phi")

    return out_dphi


def get_hpca_dist(inp, azimuth):
    r"""Returns pseudo-3D particle data structures containing mms hpca data
    for use with spd_slice2d.

    Parameters
    ----------
    inp : xarray.DataArray
        HPCA ion spec
    azimuth : xarray.DataArray
        Azimuthal angles.

    Returns
    -------
    out : dict
        HPCA distribution


    """

    # check if the time series is monotonic to avoid doing incorrect
    # calculations when there's a problem with the CDF files
    time_data = azimuth.time.data

    n_mono = np.where(time_data[1:] <= time_data[:-1])
    assert len(n_mono[0]) == 0, "non-monotonic data found in the Epoch_Angles"

    # find azimuth times with complete 1/2 spins of particle data this is
    # used to determine the number of 3D distributions that will be created
    # and where their corresponding data is located in the particle data
    # structure
    n_times = len(azimuth.data[0, 0, :, 0])
    data_idx = np.searchsorted(inp.time.data, time_data) - 1
    full = np.argwhere((data_idx[1:] - data_idx[:-1]) == n_times)
    assert len(full) != 0, "Azimuth data does not cover current time range"

    # filter times when azimuth data is all zero
    #   -just check the first energy & elevation
    #   -assume azimuth values are positive
    n_expected = azimuth.data[full, 0, 0, :].size
    n_valid_az = np.count_nonzero(azimuth.data[full, 0, 0, :])
    assert n_valid_az == n_expected, "zeroes found in the azimuth array"

    full = np.squeeze(full)

    data_idx = data_idx[full].flatten()

    # Initialize energies, angles, and support data
    # final dimensions for a single distribution (energy-azimuth-elevation)
    azimuth_dim = azimuth.data.shape
    dim = (azimuth_dim[3], azimuth_dim[2], azimuth_dim[1])

    specie = inp.attrs["FIELDNAM"].split(" ")[0].lower()

    mass, charge = mass_and_charge[specie]

    out_bins = np.ones(dim) + 1
    out_denergy = np.zeros(dim)

    energy_len = len(inp.ccomp.data)
    theta_len = len(inp.rcomp.data)
    phi_len = azimuth_dim[1]

    # energy bins are constant
    out_energy = _get_energy(inp, dim)

    out_theta = _get_theta(inp, azimuth_dim)

    out_dtheta = np.zeros([energy_len, theta_len, azimuth_dim[2]]) + 22.5

    out_phi = _get_phi(azimuth, full)

    out_dphi = _get_dphi(azimuth, full, out_phi, inp)

    out_data = np.zeros((full.size, dim[0], dim[1], dim[2]))

    # copy particle data
    for i in range(full.size):
        # need to extract the data from the center of the half-spin
        if data_idx[i] - n_times / 2.0 >= 0:
            start_idx = int(data_idx[i] - n_times / 2.0)
        else:
            continue

        if data_idx[i] + n_times / 2.0 - 1 < len(inp.time):
            end_idx = int(data_idx[i] + n_times / 2.0)
        else:
            continue

        out_data[i, ...] = np.transpose(
            inp.data[start_idx:end_idx, :, :],
            [2, 0, 1],
        )

    out = {
        "data": out_data,
        "bins": out_bins,
        "theta": out_theta,
        "phi": out_phi,
        "energy": out_energy,
        "dtheta": out_dtheta,
        "dphi": out_dphi,
        "denergy": out_denergy,
        "n_energy": energy_len,
        "n_theta": theta_len,
        "n_phi": phi_len,
        "n_times": full.size,
        "project_name": "MMS",
        "species": specie,
        "charge": charge,
        "units_name": "df_cm",
        "mass": mass,
    }

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .db_init import db_init
from .read_lfr_density import read_lfr_density
from .read_tnr import read_tnr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["db_init", "read_tnr", "read_lfr_density"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os

# 3rd party imports
import numpy as np
import pycdfpp

from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_skymap import ts_skymap
from .db_get_ts import db_get_ts

# Local imports
from .db_init import MMS_CFG_PATH
from .get_data import get_data

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def remove_edist_background(vdf, n_sec: float = 0.0, n_art: float = -1.0):
    r"""Remove secondary photoelectrons from electron distribution function
    according to [1]_.

    Parameters
    ----------
    vdf : xarray.Dataset
        Measured electron velocity distribution function.
    n_sec : float, Optional
        Artificial secondary electron density (isotropic). Default is 0.
    n_art : float, Optional
        Artificial photoelectron density (sun-angle dependant),
        Default is ephoto_scale from des-emoms GlobalAttributes.

    Returns
    -------
    vdf_new : xarray.Dataset
        Electron VDF with photoelectrons removed.
    vdf_bkg : xarray.Dataset
        Photoelectron VDF.
    photoe_scle : float
        Artificial photoelectron and secondary electron density

    References
    ----------
    .. [1]  Gershman, D. J., Avanov, L. A., Boardsen,S. A., Dorelli, J. C.,
            Gliese, U., Barrie, A. C.,... Pollock, C. J. (2017). Spacecraft
            and instrument photoelectrons measured by the dual electron
            spectrometers on MMS. Journal of Geophysical Research:Space
            Physics,122, 11,548–11,558. https://doi.org/10.1002/2017JA024518

    """

    # Time interval of VDF
    tint = list(datetime642iso8601(vdf.time.data[[0, -1]]))

    # Get spacecraft index from VDF metadata
    mms_id = vdf.data.attrs["CATDESC"].split(" ")[0].lower()
    mms_id = int(mms_id[-1])

    # Get data sample rate from VDF metadata
    if "brst" in vdf.data.attrs["FIELDNAM"].lower():
        data_rate = "brst"
        logging.info("Burst resolution data is used")
    elif "fast" in vdf.data.attrs["FIELDNAM"].lower():
        data_rate = "fast"
        logging.info("Fast resolution data is used")
    else:
        raise TypeError("Could not identify if data is fast or burst.")

    vdf_tmp = time_clip(vdf, tint)
    vdf_new = np.zeros_like(vdf_tmp.data.data)
    vdf_bkg = np.zeros_like(vdf_tmp.data.data)

    dataset_name = f"mms{mms_id}_fpi_{data_rate}_l2_des-dist"
    startdelphi_count = db_get_ts(
        dataset_name,
        f"mms{mms_id}_des_startdelphi_count_{data_rate}",
        tint,
        verbose=False,
    )

    # Load the elctron number density to get the name of the photoelectron
    # model file, and the photoelectron scaling factor
    n_e = get_data(f"ne_fpi_{data_rate}_l2", tint, mms_id, verbose=False)

    photoe_scle = n_e.attrs["GLOBAL"]["Photoelectron_model_scaling_factor"]
    photoe_scle = float(photoe_scle)

    # Load the model internal photoelectrons
    bkg_fname = n_e.attrs["GLOBAL"]["Photoelectron_model_filenames"]

    # Read the current version of the MMS configuration file
    with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
        config = json.load(fs)

    data_path = os.path.normpath(config["local"])

    bkg_fname = os.path.join(data_path, "models", "fpi", bkg_fname)

    f = pycdfpp.load(bkg_fname)

    if data_rate.lower() == "brst":
        prefs = ["mms_des_bgdist_p0", "mms_des_bgdist_p0"]
    else:
        prefs = ["mms_des_bgdist", "mms_des_bgdist"]

    vdf_bkg01 = [
        np.transpose(f[f"{prefs[0]}_{data_rate}"].values, [0, 3, 1, 2]),
        np.transpose(f[f"{prefs[1]}_{data_rate}"].values, [0, 3, 1, 2]),
    ]

    # Overwrite fraction of photoelectron if provided by user.
    if n_art > 0:
        n_photo = n_art
    else:
        n_photo = photoe_scle

    for i, _ in enumerate(vdf.time.data):
        istartdelphi_count = startdelphi_count.data[i]
        iebgdist = int(np.fix(istartdelphi_count / 16))

        esteptable_idx = int(vdf.attrs["esteptable"][i])
        vdf_bkg_tmp_data = vdf_bkg01[esteptable_idx][iebgdist, ...]
        vdf_bkg_tmp = n_photo * vdf_bkg_tmp_data

        if n_sec > 0:
            vdf_bkg_av = np.nanmean(vdf_bkg_tmp_data, axis=2)
            vdf_bkg_av = np.tile(vdf_bkg_av, (vdf_bkg_tmp_data.shape[2], 1, 1))
            vdf_bkg_av = np.transpose(vdf_bkg_av, [1, 2, 0])

            vdf_bkg_tmp += n_sec * vdf_bkg_av

        vdf_new_tmp = vdf.data.data[i, ...] - vdf_bkg_tmp
        vdf_new_tmp[vdf_new_tmp < 0] = 0.0
        vdf_bkg_tmp[vdf_bkg_tmp < 0] = 0.0
        vdf_new[i, ...] = vdf_new_tmp
        vdf_bkg[i, ...] = vdf_bkg_tmp

    # Construct the new VDFs
    glob_attrs = vdf.attrs
    vdf_attrs = vdf.data.attrs
    coords_attrs = {k: vdf[k].attrs for k in ["time", "energy", "phi", "theta"]}

    vdf_new = ts_skymap(
        vdf.time.data,
        vdf_new,
        vdf.energy.data,
        vdf.phi.data,
        vdf.theta.data,
        attrs=vdf_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )
    vdf_new.attrs = vdf.attrs
    vdf_bkg = ts_skymap(
        vdf.time.data,
        vdf_bkg,
        vdf.energy.data,
        vdf.phi.data,
        vdf.theta.data,
        attrs=vdf_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )
    vdf_bkg.attrs = vdf.attrs

    return vdf_new, vdf_bkg, photoe_scle



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from copy import deepcopy

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolosk@irf.se"
__copyright__ = "Copyright 2020-2025"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _reject_outliers(data, outliers_q=0.99):

    q = np.nanquantile(data, q=outliers_q)
    data_clean = np.where(data < q, data, np.nan)

    return data_clean


def feeps_remove_sunlit_sectors(feeps_eyes, outliers_q=0.99):
    r"""
    Remove sunlit sectors for each eye in the all eyes FEEPS data product. This is
    needed because the FEEPS contaminated sectors csv files are not updated often enough
    and there is visible spintone contamination in the data at times between updates.

    The function removes sunlit sectors from the all FEEPS eyes data product by
    identifying the sectors that are sunlit and replacing them with NaN. The idea is to
    make tables of 24 eyes x 64 sectors, similar to the ones used in the FEEPS
    contaminated sectors csv files.

    The function calculates the average observed flux for each eye-sector pair and
    places it in a 24x64 array. Then replaces with NaN the eye-sector pairs that have
    a value greater than the outliers_q percentile of the data.

    Parameters
    ----------
    feeps_eyes : xarray.Dataset
        The all FEEPS eyes data product.
    outliers_q : float, optional
        The quantile to use for outlier rejection. The default is 0.99.

    Returns
    -------
    feeps_eyes_new : xarray.Dataset
        The all FEEPS eyes data product with sunlit sectors removed.
    """

    feeps_data_vars = list(feeps_eyes.data_vars.keys())
    i_en = 3
    top_eyes_keys = [data_var for data_var in feeps_data_vars if "top" in data_var]
    bot_eyes_keys = [data_var for data_var in feeps_data_vars if "bot" in data_var]

    table = np.empty(shape=(24, 64)) * np.nan

    for key in top_eyes_keys:
        key_id = int(key.split("-")[1])

        for ss in np.arange(0, 64):

            times_in_ss = feeps_eyes.where(
                feeps_eyes.spinsectnum == ss,
                drop=True,
            ).time
            data_in_ss = feeps_eyes[key].sel(time=times_in_ss).data[:, i_en]
            data_in_ss = np.nansum(data_in_ss, axis=0)
            table[key_id - 1, ss] = data_in_ss

    for key in bot_eyes_keys:
        key_id = int(key.split("-")[1])

        for ss in np.arange(0, 64):

            times_in_ss = feeps_eyes.where(
                feeps_eyes.spinsectnum == ss,
                drop=True,
            ).time
            data_in_ss = feeps_eyes[key].sel(time=times_in_ss).data[:, i_en]
            data_in_ss = np.nansum(data_in_ss, axis=0)
            table[key_id - 1 + 12, ss] = data_in_ss

    table_clean = np.empty_like(table) * np.nan
    table_clean = _reject_outliers(table, outliers_q=outliers_q)

    table = np.where(table == 0, np.nan, table)
    table_clean = np.where(table_clean == 0, np.nan, table_clean)

    nans = np.isnan(table)
    nans_clean = np.isnan(table_clean)

    new_nans = nans_clean & ~nans

    new_nans_indices = np.argwhere(new_nans)

    new_cs = new_nans_indices + np.array([1, 0])
    feeps_eyes_new = deepcopy(feeps_eyes)

    for cs in new_cs:

        eye_id = int(cs[0])
        ss = int(cs[1])

        if eye_id <= 12:

            feeps_eyes_new[f"top-{eye_id:d}"] = xr.where(
                feeps_eyes_new["spinsectnum"] != ss,
                feeps_eyes_new[f"top-{eye_id:d}"],
                np.nan,
                keep_attrs=True,
            )
        else:
            feeps_eyes_new[f"bottom-{eye_id - 12:d}"] = xr.where(
                feeps_eyes_new["spinsectnum"] != ss,
                feeps_eyes_new[f"bottom-{eye_id - 12:d}"],
                np.nan,
                keep_attrs=True,
            )

    return feeps_eyes_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
from copy import deepcopy

# Local imports
from .feeps_correct_energies import feeps_correct_energies
from .feeps_flat_field_corrections import feeps_flat_field_corrections
from .feeps_remove_bad_data import feeps_remove_bad_data
from .feeps_remove_sun import feeps_remove_sun
from .feeps_split_integral_ch import feeps_split_integral_ch

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolosk@irf.se"
__copyright__ = "Copyright 2020-2025"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def feeps_corrections(feeps_alle):
    r"""
    Apply the FEEPS corrections to the all FEEPS eyes data product. The function
    applies the following corrections: feeps_correct_energies,
    feeps_flat_field_corrections, feeps_remove_bad_data and
    feeps_remove_sun.

    Parameters
    ----------
    feeps_alle : xarray.Dataset
        The all FEEPS eyes data product.

    Returns
    -------
    feeps_alle_clean : xarray.Dataset
        The all FEEPS eyes data product with the corrections applied.

    """

    # Make a deep copy of the input dataset to avoid modifying the original data
    feeps_alle_clean = deepcopy(feeps_alle)

    feeps_alle_clean = feeps_correct_energies(feeps_alle_clean)
    feeps_alle_clean = feeps_flat_field_corrections(feeps_alle_clean)
    feeps_alle_clean = feeps_remove_bad_data(feeps_alle_clean)

    split_int_ch = feeps_split_integral_ch(feeps_alle_clean)
    feeps_alle_clean = split_int_ch[0]
    # feeps_alle_500kev = split_int_ch[1];

    feeps_alle_clean = feeps_remove_sun(feeps_alle_clean)

    return feeps_alle_clean



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def find_closest(inp1, inp2):
    r"""Finds pairs that are closest to each other in two time series.

    Parameters
    ----------
    inp1 : ndarray
        Vector with time instants.
    inp2 : ndarray
        Vector with time instants.

    Returns
    -------
    t1new : ndarray
        Identified time instants that are closest each other.
    t2new : ndarray
        Identified time instants that are closest each other.
    ind1new : ndarray
        Identified time instants that are closest each other.
    ind2new : ndarray
        Identified time instants that are closest each other.

    """

    t1_orig = inp1
    t2_orig = inp2
    flag = True

    nt1, nt2 = [len(t) for t in [inp1, inp2]]

    while flag:
        flag_t1 = np.zeros(inp1.shape)
        tckt1 = interpolate.interp1d(
            inp1,
            np.arange(nt1),
            kind="nearest",
            fill_value="extrapolate",
        )
        flag_t1[tckt1(inp2)] = 1

        flag_t2 = np.zeros(inp2.shape)
        tckt2 = interpolate.interp1d(
            inp2,
            np.arange(nt2),
            kind="nearest",
            fill_value="extrapolate",
        )
        flag_t2[tckt2(inp1)] = 1

        ind_zeros_t1 = np.where(flag_t1 == 0)[0]
        ind_zeros_t2 = np.where(flag_t2 == 0)[0]
        if ind_zeros_t1:
            inp1 = np.delete(inp1, ind_zeros_t1)
        elif ind_zeros_t2:
            inp2 = np.delete(inp2, ind_zeros_t2)
        else:
            break

    tckt1_orig = interpolate.interp1d(t1_orig, np.arange(nt1), kind="nearest")
    tckt2_orig = interpolate.interp1d(t2_orig, np.arange(nt2), kind="nearest")

    return inp1, inp2, tckt1_orig(inp1), tckt2_orig(inp2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


# 3rd party imports
import numpy as np
from xarray.core.dataset import Dataset

# Local imports
from pyrfu.pyrf.ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def average_vdf(vdf, n_pts, method: str = "mean"):
    r"""Time averages the velocity distribution functions over `n_pts` in time.

    Parameters
    ----------
    vdf : xarray.DataArray
        Time series of the velocity distribution function.
    n_pts : int
        Number of points (samples) of the averaging window.
    method : {'mean', 'sum'}, Optional
        Method for averaging. Use 'sum' for counts. Default is 'mean'.

    Returns
    -------
    vdf_avg : xarray.DataArray
        Time series of the time averaged velocity distribution function.

    """
    # Check input type
    if not isinstance(vdf, Dataset):
        raise TypeError("vdf must be a xarray.Dataset")

    if not isinstance(n_pts, int):
        raise TypeError("n_pts must be an integer")

    if n_pts % 2 == 0:
        raise ValueError("The number of distributions to be averaged must be an odd")

    n_vdf = len(vdf.time.data)
    times = vdf.time.data

    pad_value = np.floor(n_pts / 2)
    avg_inds = np.arange(pad_value, n_vdf - pad_value, n_pts, dtype=int)
    time_avg = times[avg_inds]

    energy_avg = np.zeros((len(avg_inds), vdf.data.shape[1]))
    phi_avg = np.zeros((len(avg_inds), vdf.data.shape[2]))
    vdf_avg = np.zeros((len(avg_inds), *vdf.data.shape[1:]))

    for i, avg_ind in enumerate(avg_inds):
        l_bound = int(avg_ind - pad_value)
        r_bound = int(avg_ind + pad_value)
        if method == "mean":
            vdf_avg[i, ...] = np.nanmean(vdf.data.data[l_bound:r_bound, ...], axis=0)
        elif method == "sum":
            vdf_avg[i, ...] = np.nansum(vdf.data.data[l_bound:r_bound, ...], axis=0)
        else:
            raise NotImplementedError("method not implemented feel free to do it!!")

        energy_avg[i, ...] = np.nanmean(vdf.energy.data[l_bound:r_bound, ...], axis=0)
        phi_avg[i, ...] = np.nanmean(vdf.phi.data[l_bound:r_bound, ...], axis=0)

    vdf_avg = ts_skymap(time_avg, vdf_avg, energy_avg, phi_avg, vdf.theta.data)
    vdf_avg.attrs = vdf.attrs

    vdf_avg.time.attrs = vdf.time.attrs
    for k in vdf:
        vdf_avg[k].attrs = vdf[k].attrs

    vdf_avg.attrs["energy0"] = vdf.attrs["energy0"]
    vdf_avg.attrs["energy1"] = vdf.attrs["energy1"]
    vdf_avg.attrs["esteptable"] = vdf.attrs["esteptable"][: len(avg_inds)]
    vdf_avg.attrs["delta_energy_minus"] = vdf.attrs["delta_energy_minus"][avg_inds]
    vdf_avg.attrs["delta_energy_plus"] = vdf.attrs["delta_energy_plus"][avg_inds]

    return vdf_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from .datetime642iso8601 import datetime642iso8601

# Local imports
from .iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def extend_tint(tint, ext: list = None):
    r"""Extends time interval.

    Parameters
    ----------
    tint : list of str
        Reference time interval to extend.
    ext : list of float or list of float
        Number of seconds to extend time interval
        [left extend, right extend].

    Returns
    -------
    tint_new : list of str
        Extended time interval.

    Examples
    --------
    >>> from pyrfu import pyrf

    Time interval

    >>> tints = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]

    Load spacecraft position

    >>> tints_long = pyrf.extend_tint(tint, [-100, 100])

    """

    # Set default extension
    if ext is None:
        ext = [-60.0, 60.0]

    # Make sure tint and ext are 2 elements array_like
    message = "must be array_like with 2 elements"
    assert isinstance(tint, (np.ndarray, list)) and len(tint) == 2, f"tint {message}"
    assert isinstance(ext, (np.ndarray, list)) and len(ext) == 2, f"ext {message}"

    # Convert extension to timedelta64[ns]
    ext = (np.array(ext) * 1e9).astype("timedelta64[ns]")

    # Original time interval to datetime64[ns]
    if isinstance(tint[0], np.datetime64):
        tint_ori = tint
    elif isinstance(tint[0], str):
        tint_ori = iso86012datetime64(np.array(tint))
    else:
        raise TypeError("Invalid time format!! Must be datetime64 or str!!")

    # New time interval in iso 8601 format
    tint_new = list(datetime642iso8601(tint_ori + ext))

    return tint_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Tuple, Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def mean_field(inp: DataArray, deg: int) -> Tuple[DataArray, DataArray]:
    r"""Estimate the mean and wave fields.

    The mean field is computed by fitting a polynomial of degree `deg` to the
    input data. The wave field is then computed as the difference between the
    input data and the mean field.

    Parameters
    ----------
    inp : DataArray
        Input data.
    deg : int
        Degree of the polynomial.

    Returns
    -------
    Tuple
        Mean field and wave field.

    Raises
    ------
    TypeError
        If input is not a xarray.DataArray.

    """
    # Checking input
    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a xarray.DataArray")

    # Extracting time and data
    time: NDArray[np.datetime64] = inp.time.data
    data: NDArray[np.float64] = inp.data.astype(np.float64)  # force to double precision
    time_ints: NDArray[np.uint16] = np.arange(len(time), dtype=np.uint16)

    # Preallocating output
    inp_mean: NDArray[np.float64] = np.zeros_like(data, dtype=np.float64)
    inp_wave: NDArray[np.float64] = np.zeros_like(data, dtype=np.float64)

    for i in range(data.shape[1]):
        # Polynomial fit
        polynomial_coeffs: NDArray[np.float64] = np.polyfit(time_ints, data[:, i], deg)

        # Computing mean and wave field
        inp_mean[:, i] = np.polyval(polynomial_coeffs, time_ints)
        inp_wave[:, i] = data[:, i] - inp_mean[:, i]

    # Time series
    inp_mean_ts: DataArray = ts_vec_xyz(inp.time.data, inp_mean)
    inp_wave_ts: DataArray = ts_vec_xyz(inp.time.data, inp_wave)

    return inp_mean_ts, inp_wave_ts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from cdflib import cdfepoch

# Local imports
from .timevec2iso8601 import timevec2iso8601

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ttns2datetime64(time):
    r"""Convert time in epoch_tt2000 (nanosedconds since J2000) to datetime64
    in ns units.

    Parameters
    ----------
    time : ndarray
        Time in epoch_tt2000 (nanoseconds since J2000) format.

    Returns
    -------
    time_datetime64 : ndarray
        Time in datetime64 format in ns units.

    """

    message = "time must be float, int, or array_like"
    assert isinstance(time, (float, int, list, np.ndarray)), message

    #
    time_tt2000 = cdfepoch.breakdown_tt2000(time)

    # Convert to ISO 8601 string 'YYYY-MM-DDThh:mm:ss.mmmuuunnn'
    time_iso8601 = timevec2iso8601(time_tt2000)

    #
    time_datetime64 = time_iso8601.astype("datetime64[ns]")

    return time_datetime64



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import json
import logging
import os
import warnings
from shutil import copy, copyfileobj
from tempfile import NamedTemporaryFile
from typing import Literal, Optional, Union

# 3rd party imports
import tqdm

# Local imports
from pyrfu.mms.db_init import MMS_CFG_PATH
from pyrfu.mms.list_files_ancillary_sdc import list_files_ancillary_sdc
from pyrfu.mms.list_files_sdc import _login_lasp

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)

LASP_PUBL = "https://lasp.colorado.edu/mms/sdc/public/files/api/v1/"
LASP_SITL = "https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/"


def _make_path_local(
    file: dict,
    product: Literal["predatt", "predeph", "defatt", "defeph"],
    mms_id: Union[str, int],
    data_path: Optional[str] = "",
) -> str:
    r"""Construct path of the data file using the standard convention.

    Parameters
    ----------
    file : dict
        File information.
    product : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.
    mms_id : str or int
        Spacecraft index.
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu/mms/config.json`.

    Returns
    -------
    str
        Full path of the data file.

    Raises
    ------
    FileNotFoundError
        If the local data directory doesn't exist.

    """
    if not data_path:
        # Read the current version of the MMS configuration file
        with open(MMS_CFG_PATH, "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local"])
    else:
        data_path = os.path.normpath(data_path)

    if not os.path.exists(data_path):
        raise FileNotFoundError("local data directory doesn't exist!")

    path_list = [
        data_path,
        "ancillary",
        f"mms{mms_id}",
        product,
    ]

    return os.path.join(*path_list, file["file_name"])


def download_ancillary(
    product: Literal["predatt", "predeph", "defatt", "defeph"],
    tint: list,
    mms_id: Union[str, int],
    data_path: Optional[str] = "",
):
    r"""Download files from MMS SDC.

    Download ancillary files containing field `var_str` over the time interval `tint`
    for the spacecraft `mms_id` to `data_path`.

    Parameters
    ----------
    product : {"predatt", "predeph", "defatt", "defeph"}
        Ancillary type.
    tint : list
        Time interval
    mms_id : str or int
        Spacecraft index
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu/mms/config.json`

    """
    # List files in MMS SDC that match the request
    files_in_interval = list_files_ancillary_sdc(tint, mms_id, product)

    # Start session on MMS SDC ("public" or "sitl")
    sdc_session, headers, _ = _login_lasp()

    for file in files_in_interval:
        # Create local path following tree structure for the CDF files
        out_file = _make_path_local(file, product, mms_id, data_path)
        out_path = os.path.dirname(out_file)

        logging.info(
            "Downloading %s from %s...", os.path.basename(out_file), file["url"]
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ResourceWarning)
            fsrc = sdc_session.get(
                file["url"], stream=True, verify=True, headers=headers
            )

        with NamedTemporaryFile(delete=False) as ftmp:
            with tqdm.tqdm.wrapattr(
                fsrc.raw, "read", total=file["file_size"], ncols=60
            ) as fsrc_raw:
                with open(ftmp.name, "wb") as fs:
                    copyfileobj(fsrc_raw, fs)

        os.makedirs(out_path, exist_ok=True)

        # if the download was successful, copy to data directory
        copy(ftmp.name, out_file)
        fsrc.close()

    sdc_session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import os
import shutil
import warnings

# 3rd party imports
import cycler
import matplotlib as mpl
from matplotlib import style

# Local imports
from .add_position import add_position
from .annotate_heatmap import annotate_heatmap
from .colorbar import colorbar
from .ion_brazil_plot_thresh import ion_brazil_plot_thresh
from .make_labels import make_labels
from .mms_pl_config import mms_pl_config
from .pl_scatter_matrix import pl_scatter_matrix
from .pl_tx import pl_tx
from .plot_ang_ang import plot_ang_ang
from .plot_clines import plot_clines
from .plot_contour import plot_contour
from .plot_heatmap import plot_heatmap
from .plot_line import plot_line
from .plot_magnetosphere import plot_magnetosphere
from .plot_projection import plot_projection
from .plot_reduced_2d import plot_reduced_2d
from .plot_spectr import plot_spectr
from .plot_surf import plot_surf
from .span_tint import span_tint
from .zoom import zoom

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


__all__ = [
    "add_position",
    "annotate_heatmap",
    "colorbar",
    "ion_brazil_plot_thresh",
    "make_labels",
    "mms_pl_config",
    "pl_scatter_matrix",
    "pl_tx",
    "plot_ang_ang",
    "plot_clines",
    "plot_contour",
    "plot_heatmap",
    "plot_line",
    "plot_magnetosphere",
    "plot_projection",
    "plot_reduced_2d",
    "plot_spectr",
    "plot_surf",
    "set_color_cycle",
    "span_tint",
    "use_pyrfu_style",
    "zoom",
]
PKG_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
STYLE_SHEETS = os.path.join(PKG_PATH, "stylesheets")
style.core.USER_LIBRARY_PATHS.append(STYLE_SHEETS)
style.core.reload_library()

EXTRA_COLORS = {
    "pyrfu:bg": "#eeeeee",
    "pyrfu:fg": "#444444",
    "pyrfu:bgAlt": "#e4e4e4",
    "pyrfu:red": "#af0000",
    "pyrfu:green": "#008700",
    "pyrfu:blue": "#005f87",
    "pyrfu:yellow": "#afaf00",
    "pyrfu:orange": "#d75f00",
    "pyrfu:pink": "#d70087",
    "pyrfu:purple": "#8700af",
    "pyrfu:lightblue": "#0087af",
    "pyrfu:olive": "#5f7800",
    "on:bg": "#1b2b34",
    "on:fg": "#cdd3de",
    "on:bgAlt": "#343d46",
    "on:fgAlt": "#d8dee9",
    "on:red": "#ec5f67",
    "on:orange": "#f99157",
    "on:yellow": "#fac863",
    "on:green": "#99c794",
    "on:cyan": "#5fb3b3",
    "on:blue": "#6699cc",
    "on:pink": "#c594c5",
    "on:brown": "#ab7967",
    "series:cyan": "#54c9d1",
    "series:orange": "#eca89a",
    "series:blue": "#95bced",
    "series:olive": "#ceb776",
    "series:purple": "#d3a9ea",
    "series:green": "#9bc57f",
    "series:pink": "#f0a1ca",
    "series:turquoise": "#5fcbaa",
    "transparent": "#ffffff00",
    "mms:mms1": "#000000",
    "mms:mms2": "#d55e00",
    "mms:mms3": "#009e73",
    "mms:mms4": "#56b4e9",
    "cluster:cluster1": "#ff0000",
    "cluster:cluster2": "#008000",
    "cluster:cluster3": "#0000ff",
    "cluster:cluster4": "#00ffff",
}

mpl.colors.EXTRA_COLORS = EXTRA_COLORS
mpl.colors.colorConverter.colors.update(EXTRA_COLORS)
mpl.colormaps.register(
    name="parula",
    cmap=mpl.colors.LinearSegmentedColormap.from_list(
        "parula",
        [
            (0.0592, 0.3599, 0.8684),
            (0.078, 0.5041, 0.8385),
            (0.0232, 0.6419, 0.7914),
            (0.1802, 0.7178, 0.6425),
            (0.5301, 0.7492, 0.4662),
            (0.8186, 0.7328, 0.3499),
            (0.9956, 0.7862, 0.1968),
            (0.9764, 0.9832, 0.0539),
        ],
        N=2560,
    ),
)
mpl.colormaps.register(
    name="jet_gray",
    cmap=mpl.colors.LinearSegmentedColormap.from_list(
        "jet_gray",
        [
            (0.0, "darkblue"),
            (0.3, "cyan"),
            (0.5, "gray"),  # Midpoint changed to gray
            (0.7, "yellow"),
            (1.0, "red"),
        ],
        N=2560,
    ),
)

mpl.colormaps.register(
    name="hot_desaturated",
    cmap=mpl.colors.LinearSegmentedColormap.from_list(
        "hot_desaturated",
        [
            (0.2784, 0.2784, 0.8588),
            (0.0000, 0.0000, 0.3569),
            (0.0000, 1.0000, 1.0000),
            (0.0000, 0.5000, 0.0000),
            (1.0000, 1.0000, 0.0000),
            (1.0000, 0.3765, 0.0000),
            (0.4196, 0.0000, 0.0000),
            (0.8784, 0.2980, 0.2980),
        ],
        N=2560,
    ),
)

mpl.colormaps.register(
    name="hot_white_desaturated",
    cmap=mpl.colors.LinearSegmentedColormap.from_list(
        "hot_white_desaturated",
        [
            (1.0000, 1.0000, 1.0000),
            (0.2784, 0.2784, 0.8588),
            (0.0000, 0.0000, 0.3569),
            (0.0000, 1.0000, 1.0000),
            (0.0000, 0.5000, 0.0000),
            (1.0000, 1.0000, 0.0000),
            (1.0000, 0.3765, 0.0000),
            (0.4196, 0.0000, 0.0000),
            (0.8784, 0.2980, 0.2980),
        ],
        N=2560,
    ),
)


def set_color_cycle(pal: str = "") -> (list, str):
    r"""Sets color cycle.

    Parameters
    ----------
    pal : {"Pyrfu", "Oceanic", "Tab", ""}. Optional
      The palette to use. "Tab" provides the default matplotlib palette. Default is
      "" (resets to default palette).
    """

    if pal.lower() == "pyrfu":
        colors = [
            "pyrfu:blue",
            "pyrfu:green",
            "pyrfu:red",
            "pyrfu:fg",
            "pyrfu:orange",
            "pyrfu:purple",
            "pyrfu:yellow",
            "pyrfu:lightblue",
            "pyrfu:olive",
        ]
        cmap = "hot_white_desaturated"
    elif pal.lower() == "oceanic":
        colors = [
            "on:green",
            "on:red",
            "on:blue",
            "on:cyan",
            "on:orange",
            "on:pink",
            "on:yellow",
        ]
        cmap = "ocean"
    elif pal.lower() == "tab" or pal.lower() == "tableau" or pal.lower() == "mpl":
        colors = [
            "tab:blue",
            "tab:orange",
            "tab:green",
            "tab:red",
            "tab:purple",
            "tab:brown",
            "tab:pink",
            "tab:gray",
            "tab:olive",
            "tab:cyan",
        ]
        cmap = "Spectral_r"
    else:
        colors = [
            "series:cyan",
            "series:orange",
            "series:blue",
            "series:olive",
            "series:purple",
            "series:green",
            "series:pink",
        ]
        cmap = "jet"

    mpl.rcParams["axes.prop_cycle"] = cycler.cycler(color=colors)
    mpl.rcParams["image.cmap"] = cmap

    return colors, cmap


def use_pyrfu_style(
    name: str = "default",
    color_cycle: str = "pyrfu",
    fancy_legend: bool = False,
    usetex: bool = False,
):
    r"""Setup plot style.

    Parameters
    ----------
    name : str, Optional
       Name of the style sheet. Default is "default" (see also
       https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html).
    color_cycle : str, Optional
       Name of the color cycle to use. Default is "pyrfu".
    fancy_legend : bool, Optional
       Use matplotlib's fancy legend frame. Default is False.
    usetex : bool, Optional
       Use LaTeX installation to set text. Default is False.
       If no LaTeX installation is found, this package will fallback to usetex=False.
    """

    if usetex:
        if (
            shutil.which("pdflatex") is None
            or shutil.which("dvipng") is None
            or shutil.which("gs") is None
        ):
            warnings.warn(
                "No LaTeX installation found pyrfu.plot is falling back to usetex=False"
            )
            usetex = False

    style.use(name)
    set_color_cycle(color_cycle)
    mpl.rcParams["xtick.minor.visible"] = True
    mpl.rcParams["ytick.minor.visible"] = True
    if not usetex:
        mpl.rcParams["text.usetex"] = False
        mpl.rcParams["mathtext.sf"] = "sans"
        mpl.rcParams["mathtext.fontset"] = "dejavusans"
    else:
        mpl.rcParams["text.latex.preamble"] = "\n".join(
            [
                r"\usepackage{amsmath}",
                r"\usepackage{physics}",
                r"\usepackage{siunitx}",
                r"\usepackage{bm}",
            ]
        )
    if not fancy_legend:
        mpl.rcParams["legend.frameon"] = False
        mpl.rcParams["legend.fancybox"] = False
        mpl.rcParams["legend.framealpha"] = 0.75



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.extend_tint import extend_tint
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def probe_align_times(e_xyz, b_xyz, sc_pot, z_phase):
    r"""Returns times when field-aligned electrostatic waves can be
    characterized using interferometry techniques. The same alignment
    conditions as Graham et al., JGR, 2015 are used. Optional figure produced
    showing E_FAC, probe fields, and probe potentials to view time delays
    between electric fields aligned with B.  Currently p5-p6 are not used in
    this routine; the analysis is the same as the one used for Cluster.

    For the figure the panels are :
        * (a) B in DMPA Coordinates
        * (b) Magnitude of B in and out of the spin plane
        * (c) Angles between B and probes 1 and 3 in the spin plane
              (angle between 0 and 90 degrees)
        * (d) Spacecraft potential from probes perpendicular to B
        * (e) E fields from p1-p4 and SC for probes closely aligned with B
        * (f) E in field-aligned coordinates
        * (g) E from probes p1-p2 and p3-p4.


    Parameters
    ----------
    e_xyz : xarray.DataArray
        Electric field in DSL coordinates, brst mode.
    b_xyz : xarray.DataArray
        Magnetic field in DMPA coordinates.
    sc_pot : xarray.DataArray
        L2 Spacecraft potential data. Timing corrections are applied in this
    z_phase : xarray.DataArray
        Spacecraft phase (z_phase). Obtained from ancillary_defatt.

    Returns
    -------
    start_time1 : to fill
        Start times of intervals which satisfy the probe alignment conditions
        for probe combinates p1-p2.
    end_time1 : to fill
        End times of intervals which satisfy the probe alignment conditions
        for probe combinates p1-p2.
    start_time3 : to fill
        Start times of intervals which satisfy the probe alignment conditions
        for probe combinates p3-p4.
    end_time3 : to fill
        End times of intervals which satisfy the probe alignment conditions
        for probe combinates p3-p4.

    """

    # Correct for timing in spacecraft potential data.
    e12 = ts_scalar(
        sc_pot.time.data,
        (sc_pot.data[:, 0] - sc_pot.data[:, 1]) / 0.120,
    )
    e34 = ts_scalar(
        sc_pot.time.data,
        (sc_pot.data[:, 2] - sc_pot.data[:, 3]) / 0.120,
    )
    e56 = ts_scalar(
        sc_pot.time.data,
        (sc_pot.data[:, 4] - sc_pot.data[:, 4]) / 0.0292,
    )

    v_1 = ts_scalar(sc_pot.time.data, sc_pot.data[:, 0])
    v_3 = ts_scalar(
        sc_pot.time.data + np.timedelta64(7629, "ns"),
        sc_pot.data[:, 2],
    )
    v_5 = ts_scalar(
        sc_pot.time.data + np.timedelta64(15259, "ns"),
        sc_pot.data[:, 4],
    )

    e12.time.data += np.timedelta64(26703, "ns")
    e34.time.data += np.timedelta64(30518, "ns")
    e56.time.data += np.timedelta64(34332, "ns")

    v_1, v_3, v_5 = [resample(v, v_1) for v in [v_1, v_3, v_5]]
    e12, e34, e56 = [resample(e, v_1) for e in [e12, e34, e56]]

    v_2 = v_1 - e12 * 0.120
    v_4 = v_3 - e34 * 0.120
    v_6 = v_5 - e56 * 0.0292

    sc_pot = np.hstack(
        [v_1.data, v_2.data, v_3.data, v_4.data, v_5.data, v_6.data],
    )

    sc_pot = xr.DataArray(
        sc_pot,
        coords=[v_1.time.data, np.arange(1, 7)],
        dims=["time", "probe"],
    )

    t_limit = [sc_pot.time.data[0], sc_pot.time.data[-1]]
    t_limit = [np.datetime_as_string(time, "ns") for time in t_limit]

    t_limit_long = extend_tint(t_limit, [-10, 10])

    b_xyz = time_clip(b_xyz, t_limit_long)
    b_xyz = resample(b_xyz, sc_pot)
    e_xyz = resample(e_xyz, sc_pot)

    z_phase = time_clip(z_phase, t_limit_long)

    # Remove repeated z_phase elements
    n_ph = len(z_phase)
    no_repeat = np.ones(n_ph)

    for i in range(1, n_ph):
        if z_phase.time.data[i] > z_phase.time.data[i - 1]:
            if z_phase.data[i] < z_phase.data[i - 1]:
                z_phase.data[i:] += 360.0
        else:
            no_repeat[i] = 0

    z_phase_time = z_phase.time[no_repeat == 1]
    z_phase_data = z_phase.data[no_repeat == 1]

    z_phase = ts_scalar(z_phase_time, z_phase_data)
    z_phase = resample(z_phase, sc_pot)

    # Probe angles in DSL or whatever
    phase_p = []
    for i, j in zip([1, 7, 2, 5], [6, 6, 3, 3]):
        phase_p.append(np.deg2rad(z_phase.data) + i * np.pi / j)

    r_p = [60 * np.array([np.cos(phase), np.sin(phase)]) for phase in phase_p]

    # Calculate angles between probes and direction of B in the spin plane.
    theta_pb = [None] * 4

    for i in [0, 2]:
        theta_pb[i] = r_p[i][:, 0] * b_xyz.data[:, 0] + r_p[i][:, 1] * b_xyz.data[:, 1]
        theta_pb[i] /= np.sqrt(r_p[i][:, 0] ** 2 + r_p[i][:, 1] ** 2)
        theta_pb[i] /= np.sqrt(b_xyz[:, 0] ** 2 + b_xyz[:, 1] ** 2)
        theta_pb[i] = np.arccos(abs(theta_pb[i])) * 180 / np.pi

    theta_pb[1] = theta_pb[0]
    theta_pb[3] = theta_pb[2]

    sc_v12 = (sc_pot.data[:, 0] + sc_pot.data[:, 1]) / 2
    sc_v34 = (sc_pot.data[:, 2] + sc_pot.data[:, 3]) / 2

    e_s = [None] * 4

    e_s[0] = (sc_pot.data[:, 0] - sc_v34) * 1e3 / 60
    e_s[1] = (sc_v34 - sc_pot.data[:, 0]) * 1e3 / 60
    e_s[2] = (sc_pot.data[:, 2] - sc_v12) * 1e3 / 60
    e_s[3] = (sc_v12 - sc_pot.data[:, 2]) * 1e3 / 60

    e12 = (sc_pot.data[:, 0] - sc_pot.data[:, 1]) * 1e3 / 120
    e34 = (sc_pot.data[:, 2] - sc_pot.data[:, 3]) * 1e3 / 120

    idx_b = np.sqrt(b_xyz.data[:, 0] ** 2 + b_xyz.data[:, 1] ** 2) < abs(
        b_xyz.data[:, 2],
    )
    thresh_ang = 25.0

    for e_, theta in zip(e_s, theta_pb):
        e_[theta > thresh_ang] = np.nan
        e_[idx_b] = np.nan

    sc_v12[theta_pb[2] > thresh_ang] = np.nan
    sc_v34[theta_pb[0] > thresh_ang] = np.nan

    sc_v12[idx_b] = np.nan
    sc_v34[idx_b] = np.nan



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import string
from typing import Dict, Optional

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

# Local imports
from .. import pyrf

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


__all__ = [
    "generate_timeline",
    "generate_data",
    "generate_ts",
    "generate_spectr",
    "generate_vdf",
    "generate_defatt",
]


def generate_timeline(
    f_s: float,
    n_pts: int = 10000,
    dtype: str = "datetime64[ns]",
    ref_time: str = "2019-01-01T00:00:00.000000000",
) -> NDArray[np.datetime64]:
    r"""Generate timeline for testings

    Parameters
    ----------
    f_s : float
        Sampling frequency.
    n_pts : int
        Number of points.
    dtype : str
        Data type of the time array.
    ref_time : str
        Reference time.

    Returns
    -------
    times : numpy.ndarray
        Timeline.

    """
    ref_time = np.datetime64(ref_time)
    times = [ref_time + np.timedelta64(int(i * 1e9 / f_s), "ns") for i in range(n_pts)]
    times = np.array(times).astype(dtype)
    return times


def generate_data(n_pts: int, tensor_order: int = 0) -> NDArray[np.float32]:
    r"""Generate data (numpy.ndarray) for testings

    Parameters
    ----------
    n_pts : int
        tensor_order
    tensor_order : int
        Tensor order of the data.

    Returns
    -------
    data : numpy.ndarray
        Synthetic data.

    """

    data = np.random.random((n_pts, *([3] * tensor_order)))

    return data.astype(np.float32)


def generate_ts(
    f_s: float,
    n_pts: int,
    tensor_order: int = 0,
    attrs: Optional[Dict[str, object]] = None,
) -> DataArray:
    r"""Generate timeseries for testings

    Parameters
    ----------
    f_s : float
        Sampling frequency.
    n_pts : int
        Number of points.
    tensor_order : int
        Tensor order of the data.
    attrs : dict, Optional
        Attributes of the timeseries.

    Returns
    -------
    DataArray
        Synthetic timeseries.

    """
    if attrs is None:
        attrs = {}

    time = generate_timeline(f_s, n_pts)
    data = generate_data(n_pts, tensor_order=tensor_order)

    if tensor_order == 0:
        out = pyrf.ts_scalar(time, data, attrs=attrs)
    elif tensor_order == 1:
        out = pyrf.ts_vec_xyz(time, data, attrs=attrs)
    elif tensor_order == 2:
        out = pyrf.ts_tensor_xyz(time, data, attrs=attrs)
    else:
        coords = [time, *[np.arange(data.shape[i]) for i in range(1, tensor_order + 1)]]
        dims = ["time", *list(string.ascii_lowercase[8 : 8 + tensor_order])]
        out = xr.DataArray(data, coords=coords, dims=dims, attrs=attrs)

    out.time.attrs = {"UNITS": "I AM GROOT!!"}

    return out


def generate_spectr(
    f_s: float, n_pts: int, shape: tuple, attrs: Optional[Dict[str, object]] = None
) -> DataArray:
    r"""Generates spectrum for testings

    Parameters
    ----------
    f_s : float
        Sampling frequency.
    n_pts : int
        Number of points.
    shape : tuple
        Shape of the spectrum.
    attrs : dict
        Attributes of the spectrum.

    Returns
    -------
    DataArray
        Synthetic spectrum.


    """
    out: DataArray = pyrf.ts_spectr(
        generate_timeline(f_s, n_pts),
        np.arange(shape),
        np.random.random((n_pts, shape)),
        attrs=attrs,
    )
    return out


def generate_vdf(
    f_s: float,
    n_pts: int,
    shape: list,
    energy01: Optional[bool] = None,
    species: Optional[str] = None,
    units: Optional[str] = None,
) -> Dataset:
    r"""Generate skymap for testings

    Parameters
    ----------
    f_s : float
        Sampling frequency.
    n_pts : int
        Number of points.
    shape : list
        Shape of the vdf.
    energy01 : bool, Optional
        If True, energy is set to +0.5 for odd time steps (first part of MMS mission).
    species : str, Optional
        Species of the vdf.
    units : str, Optional
        Units of the vdf.

    Returns
    -------
    Dataset
        Synthetic vdf.

    """

    if energy01 is None:
        energy01 = False

    if species is None:
        species = "ions"

    if units is None:
        units = "s^3/cm^6"

    times = generate_timeline(f_s, n_pts)

    phi = np.arange(shape[1])
    phi = np.tile(phi, (n_pts, 1))
    theta = np.arange(shape[2])
    data = np.random.random((n_pts, *shape))

    if energy01:
        energy0 = np.linspace(0, shape[0], shape[0], endpoint=False)
        energy1 = np.linspace(0, shape[0], shape[0], endpoint=False) + 0.5
        esteptable = np.arange(n_pts, dtype=np.uint8) % 2
        energy = np.empty((n_pts, shape[0]), dtype=np.float32)
        n_energy_1 = int(np.sum(esteptable))
        n_energy_0 = n_pts - n_energy_1
        energy[esteptable == 0, :] = np.tile(energy0, (n_energy_0, 1))
        energy[esteptable == 1, :] = np.tile(energy1, (n_energy_1, 1))
    else:
        energy = np.linspace(0, shape[0], shape[0], endpoint=False, dtype=np.float32)
        energy = np.tile(energy, (n_pts, 1))
        energy0 = energy[0, :]
        energy1 = energy[1, :]
        esteptable = np.zeros(n_pts, dtype=np.uint8)

    attrs = {"UNITS": units}
    glob_attrs = {
        "species": species,
        "delta_energy_plus": np.ones((n_pts, shape[0])),
        "delta_energy_minus": np.ones((n_pts, shape[0])),
    }

    out = pyrf.ts_skymap(
        times,
        data,
        energy,
        phi,
        theta,
        energy0=energy0,
        energy1=energy1,
        esteptable=esteptable,
        attrs=attrs,
        glob_attrs=glob_attrs,
    )
    return out


def generate_defatt(f_s: float, n_pts: int) -> Dataset:
    r"""Generate defatt for testings.

    Parameters
    ----------
    f_s : float
        Sampling frequency.
    n_pts : int
        Number of points.

    Returns
    -------
    defatt : Dataset
        Defatt dataset.

    """
    z_ra = generate_ts(f_s, n_pts, 0)
    z_dec = generate_ts(f_s, n_pts, 0)
    defatt_dict = {"z_ra": z_ra, "z_dec": z_dec}
    return xr.Dataset(defatt_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"

NDArrayFloats = NDArray[Union[np.float32, np.float64]]


def trace(inp: DataArray) -> DataArray:
    r"""Computes trace of the time series of 2nd order tensors.

    Parameters
    ----------
    inp : DataArray
        Time series of the input 2nd order tensor.

    Returns
    -------
    DataArray
        Time series of the trace of the input tensor.

    Raises
    ------
    TypeError
        If inp is not a xarray.DataArray.
    ValueError
        If inp is not a time series of a tensor.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field and ion temperature

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)

    Rotate to ion temperature tensor to field aligned coordinates

    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")

    Compute scalar temperature

    >>> t_i = pyrf.trace(t_xyzfac_i)

    """

    # Check input type
    if not isinstance(inp, xr.DataArray):
        raise TypeError("inp must be a xarray.DataArray")

    # Check that inp is a tensor
    if inp.ndim != 3 or inp.shape[1] != 3 or inp.shape[2] != 3:
        raise ValueError("inp must be a time series of a tensor")

    # Get diagonal elements
    inp_xx: NDArrayFloats = inp.data[:, 0, 0]
    inp_yy: NDArrayFloats = inp.data[:, 1, 1]
    inp_zz: NDArrayFloats = inp.data[:, 2, 2]

    # Compute trace
    out_data = inp_xx + inp_yy + inp_zz

    # Construct time series
    out = ts_scalar(inp.time.data, out_data, inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def optimize_nbins_1d(x, n_min: int = 1, n_max: int = 100):
    r"""Estimates the number of bins for 1d histogram that minimizes the
    risk function in [1]_ , obtained by direct decomposition of the MISE
    following the method described in [2]_ .

    Parameters
    ----------
    x : xarray.DataArray
        Input time series
    n_min : int, Optional
        Minimum number of bins. Default is 1.
    n_max : int, Optional
        Maximum number of bins. Default is 100.

    Returns
    -------
    opt_n_x : int
        Number of bins that minimizes the cost function.

    References
    ----------
    .. [1]  Rudemo, M. (1982) Empirical Choice of Histograms and Kernel Density
            Estimators. Scandinavian Journal of Statistics, 9, 65-78.

    .. [2]  Shimazaki H. and Shinomoto S., A method for selecting the bin size
            of a time histogram Neural Computation (2007) Vol. 19(6), 1503-1527
    """

    x_min, x_max = [np.min(x.data), np.max(x.data)]

    # #of Bins
    ns_x = np.arange(n_min, n_max)

    # Bin size vector
    ds_x = (x_max - x_min) / ns_x

    cs_x = np.zeros(ds_x.shape)
    # Computation of the cost function to x and y
    for i, n_x in enumerate(ns_x):
        k_i = np.histogram(x, bins=n_x)
        # The mean and the variance are simply computed from the
        # event counts in all the bins of the 1-dimensional histogram.
        k_i = k_i[0]
        k_ = np.mean(k_i)  # Mean of event count
        v_ = np.var(k_i)  # Variance of event count
        # The cost Function
        cs_x[i] = (2 * k_ - v_) / ds_x[i] ** 2

    # Optimal Bin Size Selection
    # combination of i and j that produces the minimum cost function
    idx_min = np.argmin(cs_x)  # get the index of the min Cxy

    opt_n_x = int(ns_x[idx_min])

    return opt_n_x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import interpolate

# Local imports
from ..pyrf.normalize import normalize
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _hpca_elevations():
    anode_theta = [
        123.75000,
        101.25000,
        78.750000,
        56.250000,
        33.750000,
        11.250000,
        11.250000,
        33.750000,
        56.250000,
        78.750000,
        101.25000,
        123.75000,
        146.25000,
        168.75000,
        168.75000,
        146.25000,
    ]

    anode_theta[6:14] = [anode_val + 180.0 for anode_val in anode_theta[6:14]]

    return anode_theta


def hpca_pad(vdf, saz, aze, b_xyz, elim=None):
    r"""Computes HPCA pitch angle distribution.

    Parameters
    ----------
    vdf : xarray.DataArray
        Ion PSD or flux; [nt, npo16, ner63], looking direction
    saz : xarray.DataArray
        Start index of azimuthal angle; [nt, 1], (0 - 15)
    aze : xarray.DataArray
        Azimuthal angle per energy; [nT, naz16, npo16, ner63]
    b_xyz : xarray.DataArray
        B in dmpa coordinate
    elim : list, Optional
        [emin, emax], energy range for PAD

    Returns
    -------
    pad_spec : xarray.DataArray
        PAD spectrum

    Examples
    --------
    >>> pad_ = hpca_pad(vdf, saz, aze, elev, ien, b_xyz, elim=[500, 3000])

    """

    # 1. get data
    ien = vdf.ccomp
    elev = _hpca_elevations()

    if elim is None:
        elim = [ien.data[0], ien.data[-1]]

    # azimuthal angle #
    n_az = 16

    # 2. set tint and TLIM(dist & startaz)
    n_start_az = np.where(saz.data == 0)[0][0]
    start_time_match = aze.time.data[0].view("i8")
    start_time_match -= vdf.time.data[n_start_az].view("i8")

    if start_time_match < 1e6:
        warnings.warn("start times of aze and vdf match.", UserWarning)
    else:
        raise ValueError("start times of aze and vdf don't match!")

    n_stop_az = np.where(saz.data == 15)[0][-1]

    if (n_stop_az - n_start_az + 1) // n_az == len(aze):
        warnings.warn("stop times of aze and vdf match.", UserWarning)
    else:
        raise ValueError("stop times of aze and vdf don't match!")

    tint_ = [saz.time.data[n_start_az], saz.time.data[n_stop_az]]
    tint_ = [np.datetime_as_string(time, "ns") for time in tint_]

    vdf = time_clip(vdf, tint_)

    # 3. compute PAD
    # 3.1. pitchangle
    # Default pitch angles. 15 degree angle widths
    angle_vec = np.linspace(15, 180, 12)

    d_angle = np.median(np.diff(angle_vec)) * np.ones(len(angle_vec))
    pitch_a = angle_vec - d_angle / 2

    # 3.2.data dimension
    n_po, n_en, n_ta, n_ti = len(elev), len(ien), len(aze), len(vdf)

    assert n_ti == n_ta * n_az, "aze and vdf don't match!"

    tt_ = vdf.time.data

    # 3.3.reshape aze to aze_mat
    aze_mat = np.transpose(aze.data, [1, 0, 2, 3])  # [nt, npo, ner]
    aze_mat = np.reshape(aze_mat, [n_ti, n_po, n_en])

    elev_mat = np.tile(elev, (n_ti, n_en, 1))  # [nt, npo, ner]
    elev_mat = np.transpose(elev_mat, [0, 2, 1])
    xx_ = np.sin(np.deg2rad(elev_mat)) * np.cos(np.deg2rad(aze_mat))
    yy_ = np.sin(np.deg2rad(elev_mat)) * np.sin(np.deg2rad(aze_mat))
    zz_ = np.cos(np.deg2rad(elev_mat))

    t0_ = vdf.time.data.astype(np.int64)
    t0_start = t0_[0]
    t0_ -= t0_start
    tck_ = interpolate.interp1d(
        np.arange(0, n_en * len(t0_), n_en),
        t0_,
        fill_value="extrapolate",
    )
    t1_tt = tck_(np.arange(0, n_en * len(t0_))) + t0_start
    t1_tt = t1_tt.astype("datetime64[ns]")
    b_xyz_r = resample(b_xyz, ts_scalar(t1_tt, np.zeros(len(t1_tt))))
    b_xyz_r = normalize(b_xyz_r)

    b_x = np.transpose(np.reshape(b_xyz_r.data[:, 0], [n_en, n_ti]))
    b_x = np.transpose(np.tile(b_x, [n_po, 1, 1]), [1, 0, 2])  # [nt, npo, ner]

    b_y = np.transpose(np.reshape(b_xyz_r.data[:, 1], [n_en, n_ti]))
    b_y = np.transpose(np.tile(b_y, [n_po, 1, 1]), [1, 0, 2])  # [nt, npo, ner]

    b_z = np.transpose(np.reshape(b_xyz_r.data[:, 2], [n_en, n_ti]))
    b_z = np.transpose(np.tile(b_z, [n_po, 1, 1]), [1, 0, 2])  # [nt, npo, ner]

    theta_b = np.arccos(xx_ * b_x + yy_ * b_y + zz_ * b_z)
    theta_b = np.rad2deg(theta_b)

    # 3.5.select dist for PAD
    vdfs_ = [vdf.data.copy() for _ in range(len(angle_vec))]

    vdfs_[0][theta_b > angle_vec[0]] = np.nan

    for i, vdf_ in enumerate(vdfs_):
        vdf_[theta_b < (angle_vec[i] - d_angle[i])] = np.nan
        vdf_[theta_b > angle_vec[i]] = np.nan

    vdfs_[-1][theta_b < (angle_vec[-1] - d_angle[len(angle_vec) - 1])] = np.nan

    # [n_ti, n_po, n_en] --> [n_ti, n_en]
    vdfs_ = [np.squeeze(np.nanmean(vdf_, axis=1)) for vdf_ in vdfs_]

    # average among energy dimension
    i_elim = np.argmin(abs(ien.data - elim[0]))
    e_min = ien.data[i_elim]
    j_elim = np.argmin(abs(ien.data - elim[1]))
    e_max = ien.data[j_elim]
    messsage = f"PSD/pflux pitch angle dist. from {e_min} [eV] to {e_max} [eV]"
    warnings.warn(messsage, UserWarning)

    vdfs_ = [np.nanmean(vdf_[:, i_elim:j_elim], axis=1) for vdf_ in vdfs_]
    vdfs_ = [np.squeeze(vdf_) for vdf_ in vdfs_]  # [n_ti, n_en] --> [n_ti]
    padd_ = np.transpose(np.stack(vdfs_))  # [nt, ner, npitcha12]

    # 3.6.make spectrum
    coords = [tt_, pitch_a]
    dims = ["time", "theta"]
    pad_spec = xr.DataArray(padd_, coords=coords, dims=dims)

    return pad_spec



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Sequence

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from .calc_fs import calc_fs
from .resample import resample

__author__ = "Apostolos Kolokotronis"
__email__ = "apostolos.kolokotronis@irf.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _nan_count(inp):
    r"""Counts the number of non-NaN values in the input array at each time step.

    Parameters
    ----------
    inp : DataArray or Dataset
        Input array.

    Returns
    -------
    inp_to_counts : DataArray or Dataset
        Array with the same shape as the input array, where each element
        is the number of non-NaN values at the corresponding time step.

    """
    inp_to_counts = xr.where(np.isnan(inp), 0, 1)
    # inp_to_counts = inp.where(np.isnan(inp) == True, other=1)
    # inp_to_counts = inp_to_counts.where(np.isnan(inp) == False, other=0)

    return inp_to_counts


def nanavg_4sc(b_list: Sequence[DataArray]) -> DataArray:
    r"""Average data from 4 spacecrafts while ignoring NaN values.
    Computes the input quantity at the center of mass of the MMS
    tetrahedron. When averaging, NaN values are ignored by counting the number of
    non-NaN values at each time step.

    Parameters
    ----------
    b_list : Sequence of DataArray or Dataset
        List of the time series of the quantity for each spacecraft.

    Returns
    -------
    b_avg : DataArray or Dataset
        Time series of the input quantity a the enter of mass of the
        MMS tetrahedron.

    Raises
    ------
    TypeError
        If b_list is not a list of DataArray or Dataset

    """
    # Check input type
    if not isinstance(b_list, list):
        raise TypeError("b_list must be a list")

    b_list_r = []

    for b in b_list:
        if isinstance(b, (xr.DataArray, xr.Dataset)):
            b_list_r.append(resample(b, b_list[0], f_s=calc_fs(b_list[0])))
        else:
            raise TypeError("elements of b_list must be DataArray or Dataset")

    # b_list_r = [b.where(np.isnan(b) == False, other=0) for b in b_list_r]
    b_list_r = [xr.where(np.isnan(b), 0, b) for b in b_list_r]
    b_avg_data = np.zeros(b_list_r[0].shape)
    b_nan_denom = np.zeros(b_list_r[0].shape)

    for b in b_list_r:

        b_avg_data += b.data
        b_nan_denom += _nan_count(b).data

    if "probe" in b_list[0].attrs.keys():
        b_list[0].attrs["probe"] = "4sc_avg"
    if "mms" in b_list[0].attrs.keys():
        b_list[0].attrs["mms"] = "4sc_avg"
    if "MMS" in b_list[0].attrs.keys():
        b_list[0].attrs["MMS"] = "4sc_avg"
    if "mmsId" in b_list[0].attrs.keys():
        b_list[0].attrs["mmsId"] = "4sc_avg"

    b_avg = xr.DataArray(
        b_avg_data / b_nan_denom,
        coords=b_list_r[0].coords,
        dims=b_list_r[0].dims,
        attrs=b_list[0].attrs,
    )

    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Sequence

# 3rd party imports
import numpy as np
import xarray as xr
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.calc_fs import calc_fs
from pyrfu.pyrf.resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def avg_4sc(b_list: Sequence[DataArray]) -> DataArray:
    r"""Computes the input quantity at the center of mass of the MMS
    tetrahedron.

    Parameters
    ----------
    b_list : Sequence of DataArray or Dataset
        List of the time series of the quantity for each spacecraft.

    Returns
    -------
    b_avg : DataArray or Dataset
        Time series of the input quantity a the enter of mass of the
        MMS tetrahedron.

    Raises
    ------
    TypeError
        If b_list is not a list of DataArray or Dataset

    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu.pyrf import avg_4sc

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> b_mms = [get_data("B_gse_fgm_srvy_l2", tint, i) for i in range(1, 5)]
    >>> b_xyz = avg_4sc(b_mms)

    """
    # Check input type
    if not isinstance(b_list, list):
        raise TypeError("b_list must be a list")

    b_list_r = []

    for b in b_list:
        if isinstance(b, (xr.DataArray, xr.Dataset)):
            b_list_r.append(resample(b, b_list[0], f_s=calc_fs(b_list[0])))
        else:
            raise TypeError("elements of b_list must be DataArray or Dataset")

    b_avg_data = np.zeros(b_list_r[0].data.shape)

    # for b in b_list_r:
    #     b_avg_data += b.data

    # Average the resamples time series
    if isinstance(b_list_r[0], xr.DataArray):

        for b in b_list_r:
            b_avg_data += b.data

        b_avg = xr.DataArray(
            b_avg_data / len(b_list_r),
            coords=b_list_r[0].coords,
            dims=b_list_r[0].dims,
            attrs=b_list_r[0].attrs,
        )
    else:

        data_vars_names = list(b_list_r[0].data_vars)
        data_vars_coords = [
            list(list(b_list_r[0].data_vars[data_vars_name].coords))
            for data_vars_name in data_vars_names
        ]

        for b in b_list_r:
            b_avg_data += b.data.data

        data_vars_dict = {
            data_vars_name: b_list_r[0].data_vars[data_vars_name].data
            for data_vars_name in data_vars_names
            if data_vars_name != "data"
        }
        data_vars_dict["data"] = b_avg_data / len(b_list_r)
        # return data_vars_dict
        b_avg = xr.Dataset(
            data_vars={
                data_vars_name: (data_vars_coord, data_vars_dict[data_vars_name])
                for data_vars_name, data_vars_coord in zip(
                    data_vars_names, data_vars_coords
                )
            },
            coords=b_list_r[0].coords,
            attrs=b_list_r[0].attrs,
        )
    # return data_vars_dict
    return b_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party import
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def start(inp):
    r"""Gives the first time of the time series.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series.

    Returns
    -------
    out : float
        Value of the first time in the desired format.

    """

    # Check input type
    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"

    # Make sure this is a time series
    assert list(inp.dims)[0] == "time", "inp must be a time series"

    out = inp.time.data[0].astype(np.int64) * 1e-9

    return out



#!/usr/bin/env python

# Built-in imports
import itertools

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def optimize_nbins_2d(x, y, n_min: list = None, n_max: list = None):
    r"""Estimates the number of bins for 2d histogram that minimizes
    the risk function in [1]_ , obtained by direct decomposition of the
    MISE following the method described in [2]_ .

    Parameters
    ----------
    x : xarray.DataArray
        Input time series of the first variable.
    y : xarray.DataArray
        Input time series of the second variable.
    n_min : array_like, Optional
        Minimum number of bins for each time series. Default is [1, 1]
    n_max : array_like, Optional
        Maximum number of bins for each time series. Default is [100, 100]

    Returns
    -------
    opt_n_x : int
        Number of bins of the first variable that minimizes the cost function.
    opt_n_y : int
        Number of bins of the second variable that minimizes the cost function.

    References
    ----------
    .. [1]  Rudemo, M. (1982) Empirical Choice of Histograms and Kernel
            Density Estimators. Scandinavian Journal of Statistics, 9, 65-78.

    .. [2]  Shimazaki H. and Shinomoto S., A method for selecting the bin
            size of a time histogram Neural Computation (2007) Vol. 19(6),
            1503-1527

    """

    n_min = n_min or [1, 1]
    n_max = n_max or [100, 100]

    x_min, x_max = [np.min(x.data), np.max(x.data)]
    y_min, y_max = [np.min(y.data), np.max(y.data)]

    # #of Bins
    n_x, n_y = [np.arange(min_, max_) for min_, max_ in zip(n_min, n_max)]

    d_x = (x_max - x_min) / n_x  # Bin size vector
    d_y = (y_max - y_min) / n_y  # Bin size vector

    d_xy = [[(i, j) for j in d_y] for i in d_x]

    # matrix of bin size vector
    d_xy = np.array(d_xy, dtype=[("x", float), ("y", float)])

    # Computation of the cost function to x and y
    c_xy = np.zeros(d_xy.shape)

    for i, j in itertools.product(range(len(n_x)), range(len(n_y))):
        k_i = np.histogram2d(x, y, bins=(n_x[i], n_y[j]))
        # The mean and the variance are simply computed from the event
        # counts in all the bins of the 2-dimensional histogram.
        k_i = k_i[0]
        # The cost Function
        c_xy[i, j] = (2 * np.mean(k_i) - np.var(k_i)) / (
            (d_xy[i, j][0] * d_xy[i, j][1]) ** 2
        )

    # Optimal Bin Size Selection
    # get the index in x and y that produces the minimum cost function
    n_x = int(n_x[np.where(c_xy == np.min(c_xy))[0][0]])
    n_y = int(n_y[np.where(c_xy == np.min(c_xy))[1][0]])

    return n_x, n_y



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def sph2cart(azimuth, elevation, r):
    r"""Transform spherical to cartesian coordinates

    Parameters
    ----------
    azimuth : float or ndarray
        Azimuthal angle (phi).
    elevation : float or ndarray
        Elevation angle (theta)
    r : float or ndarray
        Radius.

    Returns
    -------
    x : float or ndarray
        Cartesian x-axis coordinates.
    y : float or ndarray
        Cartesian y-axis coordinates.
    z : float or ndarray
        Cartesian z-axis coordinates

    """

    x = r * np.sin(elevation) * np.cos(azimuth)
    y = r * np.sin(elevation) * np.sin(azimuth)
    z = r * np.cos(elevation)

    return x, y, z



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.normalize import normalize
from ..pyrf.resample import resample
from ..pyrf.time_clip import time_clip

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def get_pitch_angle_dist(vdf, b_xyz, tint: list = None, verbose=True, **kwargs):
    r"""Computes the pitch angle distributions from l1b brst particle data.

    Parameters
    ----------
    vdf : xarray.Dataset
        to fill
    b_xyz : xarray.DataArray
        to fill
    tint : list of str, Optional
        Time interval for closeup.

    Returns
    -------
    pad : xarray.DataArray
        Particle pitch angle distribution

    Other Parameters
    ----------------
    angles : int or float or list of ndarray
        User defined angles.
    meanorsum : {'mean', 'sum', 'sum_weighted'}
        Method.

    Examples
    --------
    >>> from pyrfu import mms

    Define time intervals

    >>> tint_long = ["2017-07-24T12:48:34.000", "2017-07-24T12:58:20.000"]
    >>> tint_zoom = ["2017-07-24T12:49:18.000", "2017-07-24T12:49:30.000"]

    Load ions velocity distribution for MMS1

    >>> vdf_i = mms.get_data("pdi_fpi_brst_l2", tint_long, 1)

    Load magnetic field in the spacecraft coordinates system.

    >>> b_dmpa = mms.get_data("b_dmpa_fgm_brst_l2", tint_long, 1)

    Compute pitch angle distribution

    >>> options = dict(angles=24)
    >>> pad_i = mms.get_pitch_angle_dist(vdf, b_dmpa, tint_zoom, **options)

    """

    # Default pitch angles. 15 degree angle widths
    angles_v = np.linspace(15, 180, int(180 / 15))
    d_angles = np.median(np.diff(angles_v)) * np.ones(len(angles_v))

    if "angles" in kwargs:
        if isinstance(kwargs["angles"], (int, float)):
            n_angles = int(kwargs["angles"])  # Make sure input is integer
            d_angles = 180 / n_angles
            angles_v = np.linspace(d_angles, 180, n_angles)
            d_angles = d_angles * np.ones(n_angles)

            if verbose:
                logging.info("User defined number of pitch angles.")

        elif isinstance(kwargs["angles"], (list, np.ndarray)):
            angles_v = kwargs["angles"]
            d_angles = np.diff(angles_v)
            angles_v = angles_v[1:]

            if verbose:
                logging.info("User defined pitch angle limits.")

        else:
            raise ValueError("angles parameter not understood.")

    # Method
    mean_or_sum = kwargs.get("meanorsum", "mean")
    assert mean_or_sum in ["mean", "sum"], "meanorsum param. not understood."

    pitch_angles = angles_v - d_angles / 2
    n_angles = len(angles_v)

    time = vdf.time.data

    vdf0 = vdf.data.copy()

    if vdf.phi.data.ndim == 1:
        phi = np.tile(vdf.phi.data, (len(time), 1))
        phi = xr.DataArray(
            phi,
            coords=[time, np.arange(len(phi))],
            dims=["time", "idx1"],
        )
    else:
        phi = vdf.phi

    theta = vdf.theta

    if "energy0" in vdf.attrs.keys() and "energy1" in vdf.attrs.keys():
        energy0, _ = [vdf.attrs[f"energy{i}"] for i in range(2)]
    else:
        energy0, _ = vdf.energy.data[:2, :]

    if tint is not None:
        b_xyz = time_clip(b_xyz, tint)
        vdf0 = time_clip(vdf0, tint)
        phi = time_clip(phi, tint)

    time = vdf0.time.data

    # Check size of energy
    n_en, n_phi, n_theta = [len(energy0), len(phi.data[0, :]), len(theta)]

    b_xyz = resample(b_xyz, vdf0)
    b_vec = normalize(b_xyz)

    b_vec_x = np.transpose(
        np.tile(b_vec.data[:, 0], [n_en, n_phi, n_theta, 1]),
        [3, 0, 1, 2],
    )
    b_vec_y = np.transpose(
        np.tile(b_vec.data[:, 1], [n_en, n_phi, n_theta, 1]),
        [3, 0, 1, 2],
    )
    b_vec_z = np.transpose(
        np.tile(b_vec.data[:, 2], [n_en, n_phi, n_theta, 1]),
        [3, 0, 1, 2],
    )

    x_vec = np.zeros((len(time), n_phi, n_theta))
    y_vec = np.zeros((len(time), n_phi, n_theta))
    z_vec = np.zeros((len(time), n_phi, n_theta))

    for i in range(len(time)):
        x_vec[i, ...] = np.dot(
            -np.cos(np.deg2rad(phi.data[i, None])).T,
            np.sin(np.deg2rad(theta.data[:, None])).T,
        )
        y_vec[i, ...] = np.dot(
            -np.sin(np.deg2rad(phi.data[i, None])).T,
            np.sin(np.deg2rad(theta.data[:, None])).T,
        )
        z_vec[i, ...] = np.dot(
            -np.ones((n_phi, 1)),
            np.cos(np.deg2rad(theta.data[:, None])).T,
        )

    if tint is not None:
        energy = time_clip(vdf.energy, tint).data
    else:
        energy = vdf.energy.data

    x_mat = np.squeeze(
        np.transpose(np.tile(x_vec, [n_en, 1, 1, 1]), [1, 0, 2, 3]),
    )
    y_mat = np.squeeze(
        np.transpose(np.tile(y_vec, [n_en, 1, 1, 1]), [1, 0, 2, 3]),
    )
    z_mat = np.squeeze(
        np.transpose(np.tile(z_vec, [n_en, 1, 1, 1]), [1, 0, 2, 3]),
    )

    theta_b = np.rad2deg(
        np.arccos(
            x_mat * np.squeeze(b_vec_x)
            + y_mat * np.squeeze(b_vec_y)
            + z_mat * np.squeeze(b_vec_z),
        ),
    )

    dists = [vdf0.data.copy() for _ in range(n_angles)]

    pad_arr = [None] * n_angles

    for i in range(n_angles):
        dists[i][theta_b < angles_v[i] - d_angles[i]] = np.nan
        dists[i][theta_b > angles_v[i]] = np.nan

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            if mean_or_sum == "mean":
                pad_arr[i] = np.squeeze(
                    np.nanmean(np.nanmean(dists[i], axis=3), axis=2),
                )
            elif mean_or_sum == "sum":
                pad_arr[i] = np.squeeze(
                    np.nansum(np.nansum(dists[i], axis=3), axis=2),
                )
            else:
                raise ValueError("Invalid method")

    pad_arr = np.transpose(np.stack(pad_arr), [1, 0, 2])

    energy = np.mean(energy[:2, :], axis=0)

    pad = xr.Dataset(
        {
            "data": (
                ["time", "idx0", "idx1"],
                np.transpose(pad_arr, [0, 2, 1]),
            ),
            "energy": (["time", "idx0"], np.tile(energy, (len(pad_arr), 1))),
            "theta": (
                ["time", "idx1"],
                np.tile(pitch_angles, (len(pad_arr), 1)),
            ),
            "time": time,
            "idx0": np.arange(len(energy)),
            "idx1": np.arange(len(pitch_angles)),
        },
    )

    pad.attrs = vdf.attrs
    pad.attrs["mean_or_sum"] = mean_or_sum
    pad.attrs["delta_pitchangle_minus"] = d_angles * 0.5
    pad.attrs["delta_pitchangle_plus"] = d_angles * 0.5

    pad.time.attrs = vdf.time.attrs
    pad.energy.attrs = vdf.energy.attrs
    pad.data.attrs["UNITS"] = vdf.data.attrs["UNITS"]

    return pad



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from xarray.core.dataarray import DataArray

# Local imports
from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2024"
__license__ = "MIT"
__version__ = "2.4.13"
__status__ = "Prototype"


def dot(inp1: DataArray, inp2: DataArray) -> DataArray:
    r"""Compute dot product of two fields.

    Parameters
    ----------
    inp1 : DataArray
        Time series of the first field X.
    inp2 : DataArray
        Time series of the second field Y.

    Returns
    -------
    out : DataArray
        Time series of the dot product Z = X.Y.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Load magnetic field, electric field and spacecraft position

    >>> r_mms, b_mms, e_mms = [[] * 4 for _ in range(3)]
    >>> for mms_id in range(1, 5):
    >>>		r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>		e_mms.append(mms.get_data("E_gse_edp_fast_l2", tint, mms_id))

    Compute current density using curlometer technique

    >>> j_xyz, _, _, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    Compute the electric at the center of mass of the tetrahedron

    >>> e_xyz = pyrf.avg_4sc(e_mms)

    Compute J.E dissipation

    >>> je = pyrf.dot(j_xyz, e_xyz)

    """

    # Resample to first input sampling
    inp2 = resample(inp2, inp1)

    # Compute scalar product
    out_data = np.sum(inp1.data * inp2.data, axis=1)

    # Output to xarray
    out = ts_scalar(inp1.time.data, out_data)

    return out
